from __future__ import annotations

from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.config.settings import get_settings
from app.db.models import DashboardSettings

_SETTINGS_ID = 1


class SettingsRepository:
    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def get_or_create(self) -> DashboardSettings:
        existing = await self._session.get(DashboardSettings, _SETTINGS_ID)
        if existing is not None:
            return existing

        row = DashboardSettings(
            id=_SETTINGS_ID,
            sticky_threads_enabled=False,
            upstream_stream_transport="default",
            prefer_earlier_reset_accounts=False,
            routing_strategy="capacity_weighted",
            openai_cache_affinity_max_age_seconds=get_settings().openai_cache_affinity_max_age_seconds,
            import_without_overwrite=True,
            totp_required_on_login=False,
            password_hash=None,
            api_key_auth_enabled=False,
            totp_secret_encrypted=None,
            totp_last_verified_step=None,
        )
        self._session.add(row)
        try:
            await self._session.commit()
        except IntegrityError:
            await self._session.rollback()
            existing = await self._session.get(DashboardSettings, _SETTINGS_ID)
            if existing is None:
                raise
            return existing
        await self._session.refresh(row)
        return row

    async def update(
        self,
        *,
        sticky_threads_enabled: bool | None = None,
        upstream_stream_transport: str | None = None,
        prefer_earlier_reset_accounts: bool | None = None,
        routing_strategy: str | None = None,
        openai_cache_affinity_max_age_seconds: int | None = None,
        http_responses_session_bridge_prompt_cache_idle_ttl_seconds: int | None = None,
        http_responses_session_bridge_gateway_safe_mode: bool | None = None,
        sticky_reallocation_budget_threshold_pct: float | None = None,
        import_without_overwrite: bool | None = None,
        totp_required_on_login: bool | None = None,
        api_key_auth_enabled: bool | None = None,
    ) -> DashboardSettings:
        settings = await self.get_or_create()
        if sticky_threads_enabled is not None:
            settings.sticky_threads_enabled = sticky_threads_enabled
        if upstream_stream_transport is not None:
            settings.upstream_stream_transport = upstream_stream_transport
        if prefer_earlier_reset_accounts is not None:
            settings.prefer_earlier_reset_accounts = prefer_earlier_reset_accounts
        if routing_strategy is not None:
            settings.routing_strategy = routing_strategy
        if openai_cache_affinity_max_age_seconds is not None:
            settings.openai_cache_affinity_max_age_seconds = openai_cache_affinity_max_age_seconds
        if http_responses_session_bridge_prompt_cache_idle_ttl_seconds is not None:
            settings.http_responses_session_bridge_prompt_cache_idle_ttl_seconds = (
                http_responses_session_bridge_prompt_cache_idle_ttl_seconds
            )
        if http_responses_session_bridge_gateway_safe_mode is not None:
            settings.http_responses_session_bridge_gateway_safe_mode = http_responses_session_bridge_gateway_safe_mode
        if sticky_reallocation_budget_threshold_pct is not None:
            settings.sticky_reallocation_budget_threshold_pct = sticky_reallocation_budget_threshold_pct
        if import_without_overwrite is not None:
            settings.import_without_overwrite = import_without_overwrite
        if totp_required_on_login is not None:
            settings.totp_required_on_login = totp_required_on_login
        if api_key_auth_enabled is not None:
            settings.api_key_auth_enabled = api_key_auth_enabled
        await self.commit_refresh(settings)
        return settings

    async def commit_refresh(self, settings: DashboardSettings) -> None:
        await self._session.commit()
        await self._session.refresh(settings)



from __future__ import annotations

import pytest


@pytest.mark.asyncio
@pytest.mark.e2e
async def test_all_health_endpoints_respond(client):
    response = await client.get("/health")
    assert response.status_code == 200
    assert response.json() == {"status": "ok"}

    response = await client.get("/health/live")
    assert response.status_code == 200
    live_payload = response.json()
    assert live_payload["status"] == "ok"
    assert live_payload["checks"] is None
    assert live_payload.get("bridge_ring") is None

    response = await client.get("/health/ready")
    assert response.status_code == 200
    ready_payload = response.json()
    assert ready_payload["status"] == "ok"
    assert "checks" in ready_payload
    assert ready_payload["checks"]["database"] == "ok"

    response = await client.get("/health/startup")
    assert response.status_code in (200, 503)



from __future__ import annotations

import time
from dataclasses import replace

import pytest

from app.core.openai.model_registry import ModelRegistry, ReasoningLevel, UpstreamModel

pytestmark = pytest.mark.unit


def _model(slug: str) -> UpstreamModel:
    return UpstreamModel(
        slug=slug,
        display_name=slug,
        description=f"Model {slug}",
        context_window=128000,
        input_modalities=("text",),
        supported_reasoning_levels=(ReasoningLevel(effort="medium", description="balanced"),),
        default_reasoning_level="medium",
        supports_reasoning_summaries=False,
        support_verbosity=False,
        default_verbosity=None,
        prefer_websockets=False,
        supports_parallel_tool_calls=True,
        supported_in_api=True,
        minimal_client_version=None,
        priority=0,
        available_in_plans=frozenset(),
        raw={},
    )


@pytest.mark.asyncio
async def test_initial_snapshot_is_none():
    registry = ModelRegistry(ttl_seconds=60.0)
    assert registry.get_snapshot() is None


@pytest.mark.asyncio
async def test_plan_types_for_model_returns_none_when_uninitialized():
    registry = ModelRegistry(ttl_seconds=60.0)
    result = registry.plan_types_for_model("some-model")
    assert result is None


@pytest.mark.asyncio
async def test_plan_types_for_model_returns_empty_for_unknown_model():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update({"plus": [_model("model-a")]})
    result = registry.plan_types_for_model("unknown-model")
    assert result == frozenset()


@pytest.mark.asyncio
async def test_plan_types_for_model_returns_plans():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update(
        {
            "plus": [_model("model-a"), _model("model-b")],
            "pro": [_model("model-a"), _model("model-c")],
        }
    )

    assert registry.plan_types_for_model("model-a") == frozenset({"plus", "pro"})
    assert registry.plan_types_for_model("model-b") == frozenset({"plus"})
    assert registry.plan_types_for_model("model-c") == frozenset({"pro"})


@pytest.mark.asyncio
async def test_prefers_websockets_uses_snapshot_value():
    registry = ModelRegistry(ttl_seconds=60.0)
    preferred = replace(_model("model-ws"), prefer_websockets=True)
    await registry.update({"plus": [preferred]})

    assert registry.prefers_websockets("model-ws") is True
    assert registry.prefers_websockets("unknown-model") is False


def test_prefers_websockets_uses_bootstrap_fallback_when_uninitialized():
    registry = ModelRegistry(ttl_seconds=60.0)

    assert registry.prefers_websockets("gpt-5.4") is True
    assert registry.prefers_websockets("gpt-5.4-2026") is True
    assert registry.prefers_websockets("gpt-5.1") is False


@pytest.mark.asyncio
async def test_update_merges_models_across_plans():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update(
        {
            "plus": [_model("shared"), _model("plus-only")],
            "pro": [_model("shared"), _model("pro-only")],
        }
    )

    snapshot = registry.get_snapshot()
    assert snapshot is not None
    assert set(snapshot.models.keys()) == {"shared", "plus-only", "pro-only"}
    assert snapshot.plan_models["plus"] == frozenset({"shared", "plus-only"})
    assert snapshot.plan_models["pro"] == frozenset({"shared", "pro-only"})


@pytest.mark.asyncio
async def test_partial_update_preserves_stale_plans():
    registry = ModelRegistry(ttl_seconds=60.0)

    # First full update with both plans
    await registry.update(
        {
            "plus": [_model("shared"), _model("plus-only")],
            "pro": [_model("shared"), _model("pro-only")],
        }
    )

    # Partial update: only plus succeeds, pro fails (not in per_plan_results)
    await registry.update(
        {
            "plus": [_model("shared"), _model("plus-new")],
        }
    )

    snapshot = registry.get_snapshot()
    assert snapshot is not None

    # pro-only should be preserved from previous snapshot
    assert "pro-only" in snapshot.models
    assert "pro" in snapshot.model_plans.get("pro-only", frozenset())

    # plus-only should be gone (not in new plus results)
    assert "plus-only" not in snapshot.models

    # plus-new should be present
    assert "plus-new" in snapshot.models
    assert "plus" in snapshot.model_plans["plus-new"]


def test_needs_refresh_true_initially():
    registry = ModelRegistry(ttl_seconds=60.0)
    assert registry.needs_refresh() is True


@pytest.mark.asyncio
async def test_needs_refresh_false_after_update():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update({"plus": [_model("a")]})
    assert registry.needs_refresh() is False


@pytest.mark.asyncio
async def test_needs_refresh_true_after_ttl(monkeypatch):
    registry = ModelRegistry(ttl_seconds=1.0)
    await registry.update({"plus": [_model("a")]})
    assert registry.needs_refresh() is False

    # Simulate time passage by adjusting fetched_at
    snapshot = registry.get_snapshot()
    assert snapshot is not None
    snapshot.fetched_at = time.monotonic() - 2.0
    assert registry.needs_refresh() is True


@pytest.mark.asyncio
async def test_empty_update_is_noop():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update({})
    assert registry.get_snapshot() is None


def test_ttl_must_be_positive():
    with pytest.raises(ValueError, match="positive"):
        ModelRegistry(ttl_seconds=0)
    with pytest.raises(ValueError, match="positive"):
        ModelRegistry(ttl_seconds=-1.0)


@pytest.mark.asyncio
async def test_plan_models_reverse_index():
    registry = ModelRegistry(ttl_seconds=60.0)
    await registry.update(
        {
            "plus": [_model("a"), _model("b")],
            "pro": [_model("b"), _model("c")],
        }
    )

    snapshot = registry.get_snapshot()
    assert snapshot is not None
    assert snapshot.plan_models["plus"] == frozenset({"a", "b"})
    assert snapshot.plan_models["pro"] == frozenset({"b", "c"})



from __future__ import annotations

from fastapi import APIRouter, Body, Depends
from fastapi.responses import JSONResponse

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.clients.oauth import OAuthError
from app.core.errors import dashboard_error
from app.dependencies import OauthContext, get_oauth_context
from app.modules.oauth.schemas import (
    ManualCallbackRequest,
    ManualCallbackResponse,
    OauthCompleteRequest,
    OauthCompleteResponse,
    OauthStartRequest,
    OauthStartResponse,
    OauthStatusResponse,
)

router = APIRouter(
    prefix="/api/oauth",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.post("/start", response_model=OauthStartResponse)
async def start_oauth(
    request: OauthStartRequest,
    context: OauthContext = Depends(get_oauth_context),
) -> OauthStartResponse | JSONResponse:
    try:
        return await context.service.start_oauth(request)
    except OAuthError as exc:
        return JSONResponse(
            status_code=502,
            content=dashboard_error(exc.code, exc.message),
        )
    except NotImplementedError:
        return JSONResponse(
            status_code=501,
            content=dashboard_error("not_implemented", "OAuth start is not implemented"),
        )


@router.get("/status", response_model=OauthStatusResponse)
async def oauth_status(
    context: OauthContext = Depends(get_oauth_context),
) -> OauthStatusResponse | JSONResponse:
    return await context.service.oauth_status()


@router.post("/complete", response_model=OauthCompleteResponse)
async def complete_oauth(
    request: OauthCompleteRequest | None = Body(default=None),
    context: OauthContext = Depends(get_oauth_context),
) -> OauthCompleteResponse | JSONResponse:
    try:
        return await context.service.complete_oauth(request)
    except NotImplementedError:
        return JSONResponse(
            status_code=501,
            content=dashboard_error("not_implemented", "OAuth complete is not implemented"),
        )


@router.post("/manual-callback", response_model=ManualCallbackResponse)
async def manual_callback(
    request: ManualCallbackRequest,
    context: OauthContext = Depends(get_oauth_context),
) -> ManualCallbackResponse | JSONResponse:
    try:
        return await context.service.manual_callback(request.callback_url)
    except Exception as exc:
        return JSONResponse(
            status_code=500,
            content=dashboard_error("manual_callback_failed", str(exc)),
        )



from __future__ import annotations

import json
import os
import signal
import subprocess
import sys
import time
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Callable, Final
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from app.core.config.settings import get_settings

DEFAULT_STARTUP_TIMEOUT_SECONDS: Final[float] = 30.0
DEFAULT_SHUTDOWN_TIMEOUT_SECONDS: Final[float] = 15.0


@dataclass(frozen=True, slots=True)
class ServeOptions:
    host: str
    port: int
    ssl_certfile: str | None
    ssl_keyfile: str | None


@dataclass(frozen=True, slots=True)
class RuntimeMetadata:
    pid: int
    host: str
    port: int
    log_file: str


def default_runtime_dir() -> Path:
    return get_settings().encryption_key_file.parent


def default_pid_file() -> Path:
    return default_runtime_dir() / "server.pid"


def default_log_file() -> Path:
    return default_runtime_dir() / "server.log"


def build_serve_command(python_executable: str, options: ServeOptions) -> list[str]:
    command = [
        python_executable,
        "-m",
        "app.cli",
        "serve",
        "--host",
        options.host,
        "--port",
        str(options.port),
    ]
    if options.ssl_certfile:
        command.extend(["--ssl-certfile", options.ssl_certfile])
    if options.ssl_keyfile:
        command.extend(["--ssl-keyfile", options.ssl_keyfile])
    return command


def read_runtime_metadata(pid_file: Path) -> RuntimeMetadata | None:
    resolved = pid_file.expanduser()
    if not resolved.exists():
        return None
    payload = json.loads(resolved.read_text(encoding="utf-8"))
    return RuntimeMetadata(
        pid=int(payload["pid"]),
        host=str(payload["host"]),
        port=int(payload["port"]),
        log_file=str(payload["log_file"]),
    )


def write_runtime_metadata(pid_file: Path, metadata: RuntimeMetadata) -> None:
    resolved = pid_file.expanduser()
    resolved.parent.mkdir(parents=True, exist_ok=True)
    resolved.write_text(json.dumps(asdict(metadata), indent=2, sort_keys=True), encoding="utf-8")


def remove_runtime_metadata(pid_file: Path) -> None:
    pid_file.expanduser().unlink(missing_ok=True)


def is_process_running(pid: int) -> bool:
    if pid <= 0:
        return False
    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        return False
    except PermissionError:
        return True
    return True


def load_running_metadata(pid_file: Path) -> tuple[RuntimeMetadata | None, bool]:
    metadata = read_runtime_metadata(pid_file)
    if metadata is None:
        return None, False
    if is_process_running(metadata.pid):
        return metadata, False
    remove_runtime_metadata(pid_file)
    return None, True


def start_background_server(
    options: ServeOptions,
    *,
    pid_file: Path,
    log_file: Path,
    startup_timeout_seconds: float = DEFAULT_STARTUP_TIMEOUT_SECONDS,
) -> RuntimeMetadata:
    existing, stale = load_running_metadata(pid_file)
    if existing is not None:
        raise RuntimeError(
            f"background server already running with pid {existing.pid} "
            f"on {existing.host}:{existing.port} (pid file: {pid_file.expanduser()})"
        )
    if stale:
        remove_runtime_metadata(pid_file)

    resolved_pid_file = pid_file.expanduser()
    resolved_log_file = log_file.expanduser()
    resolved_pid_file.parent.mkdir(parents=True, exist_ok=True)
    resolved_log_file.parent.mkdir(parents=True, exist_ok=True)

    env = dict(os.environ)
    env["PYTHONUNBUFFERED"] = "1"

    if os.name == "nt":
        creationflags = getattr(subprocess, "CREATE_NEW_PROCESS_GROUP", 0) | getattr(subprocess, "DETACHED_PROCESS", 0)

    with resolved_log_file.open("ab") as log_handle:
        if os.name == "nt":
            process = subprocess.Popen(
                build_serve_command(sys.executable, options),
                stdin=subprocess.DEVNULL,
                stdout=log_handle,
                stderr=subprocess.STDOUT,
                cwd=os.getcwd(),
                env=env,
                creationflags=creationflags,
            )
        else:
            process = subprocess.Popen(
                build_serve_command(sys.executable, options),
                stdin=subprocess.DEVNULL,
                stdout=log_handle,
                stderr=subprocess.STDOUT,
                cwd=os.getcwd(),
                env=env,
                start_new_session=True,
            )

    metadata = RuntimeMetadata(
        pid=process.pid,
        host=options.host,
        port=options.port,
        log_file=str(resolved_log_file),
    )
    write_runtime_metadata(resolved_pid_file, metadata)

    if not wait_for_server_ready(
        metadata,
        timeout_seconds=startup_timeout_seconds,
        poll_process=lambda: process.poll(),
    ):
        terminate_process(metadata.pid)
        remove_runtime_metadata(resolved_pid_file)
        raise RuntimeError(f"background server failed readiness checks; see log file {resolved_log_file}")

    return metadata


def wait_for_server_ready(
    metadata: RuntimeMetadata,
    *,
    timeout_seconds: float,
    poll_process: Callable[[], int | None],
) -> bool:
    deadline = time.monotonic() + timeout_seconds
    health_url = _healthcheck_url(metadata.host, metadata.port)
    while time.monotonic() < deadline:
        if poll_process() is not None:
            return False
        if healthcheck_ready(health_url):
            return True
        time.sleep(0.2)
    return False


def healthcheck_ready(url: str) -> bool:
    try:
        with urlopen(url, timeout=0.5) as response:  # noqa: S310 - local readiness probe only
            return response.status == 200
    except (HTTPError, URLError, TimeoutError):
        return False


def shutdown_background_server(
    pid_file: Path,
    *,
    timeout_seconds: float = DEFAULT_SHUTDOWN_TIMEOUT_SECONDS,
) -> RuntimeMetadata | None:
    metadata, stale = load_running_metadata(pid_file)
    if metadata is None:
        if stale:
            return None
        return None

    terminate_process(metadata.pid)
    deadline = time.monotonic() + timeout_seconds
    while time.monotonic() < deadline:
        if not is_process_running(metadata.pid):
            remove_runtime_metadata(pid_file)
            return metadata
        time.sleep(0.2)

    raise RuntimeError(f"timed out waiting for pid {metadata.pid} to stop (pid file: {pid_file.expanduser()})")


def terminate_process(pid: int) -> None:
    try:
        os.kill(pid, signal.SIGTERM)
    except ProcessLookupError:
        return


def _healthcheck_url(host: str, port: int) -> str:
    probe_host = host
    if host in {"0.0.0.0", ""}:
        probe_host = "127.0.0.1"
    elif host == "::":
        probe_host = "::1"

    if ":" in probe_host and not probe_host.startswith("["):
        probe_host = f"[{probe_host}]"
    return f"http://{probe_host}:{port}/health/live"



"""add api key account assignments

Revision ID: 20260406_000000_add_api_key_account_assignments
Revises: 20260403_000000_add_credit_api_key_limit_values
Create Date: 2026-04-06
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260406_000000_add_api_key_account_assignments"
down_revision = "20260403_000000_add_credit_api_key_limit_values"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _indexes(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(index["name"]) for index in inspector.get_indexes(table_name) if index.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_key_accounts"):
        op.create_table(
            "api_key_accounts",
            sa.Column("api_key_id", sa.String(), sa.ForeignKey("api_keys.id", ondelete="CASCADE"), nullable=False),
            sa.Column("account_id", sa.String(), sa.ForeignKey("accounts.id", ondelete="CASCADE"), nullable=False),
            sa.Column("created_at", sa.DateTime(), server_default=sa.func.now(), nullable=False),
            sa.PrimaryKeyConstraint("api_key_id", "account_id"),
        )

    existing_indexes = _indexes(bind, "api_key_accounts")
    if "idx_api_key_accounts_account_id" not in existing_indexes:
        op.create_index("idx_api_key_accounts_account_id", "api_key_accounts", ["account_id"], unique=False)


def downgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_key_accounts"):
        return

    existing_indexes = _indexes(bind, "api_key_accounts")
    if "idx_api_key_accounts_account_id" in existing_indexes:
        op.drop_index("idx_api_key_accounts_account_id", table_name="api_key_accounts")
    op.drop_table("api_key_accounts")



from __future__ import annotations

import time
from dataclasses import dataclass

import anyio


@dataclass(slots=True)
class _CachedFirewallDecision:
    allowed: bool
    expires_at: float


class FirewallIPCache:
    def __init__(self, ttl_seconds: int = 15) -> None:
        if ttl_seconds <= 0:
            raise ValueError("ttl_seconds must be positive")
        self._ttl_seconds = ttl_seconds
        self._cache: dict[str, _CachedFirewallDecision] = {}
        self._lock = anyio.Lock()
        self._version = 0

    @property
    def version(self) -> int:
        return self._version

    async def is_allowed(self, ip: str) -> bool | None:
        entry = self._cache.get(ip)
        if entry is None:
            return None
        if time.monotonic() >= entry.expires_at:
            return None
        return entry.allowed

    async def set(self, ip: str, allowed: bool, *, if_version: int | None = None) -> None:
        async with self._lock:
            if if_version is not None and if_version != self._version:
                return
            self._cache[ip] = _CachedFirewallDecision(
                allowed=allowed,
                expires_at=time.monotonic() + self._ttl_seconds,
            )

    def invalidate_all(self) -> None:
        self._cache.clear()
        self._version += 1


_firewall_ip_cache = FirewallIPCache(ttl_seconds=2)


def get_firewall_ip_cache() -> FirewallIPCache:
    return _firewall_ip_cache



"""normalize sqlite account status casing

Revision ID: 20260319_183000_normalize_sqlite_account_status_casing
Revises: 20260319_111500_merge_prompt_cache_ttl_and_upstream_transport_heads
Create Date: 2026-03-19
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "20260319_183000_normalize_sqlite_account_status_casing"
down_revision = "20260319_111500_merge_prompt_cache_ttl_and_upstream_transport_heads"
branch_labels = None
depends_on = None

_ACCOUNT_STATUS_RENAMES: tuple[tuple[str, str], ...] = (
    ("ACTIVE", "active"),
    ("RATE_LIMITED", "rate_limited"),
    ("QUOTA_EXCEEDED", "quota_exceeded"),
    ("PAUSED", "paused"),
    ("DEACTIVATED", "deactivated"),
)


def _apply_status_renames(bind: sa.engine.Connection, renames: tuple[tuple[str, str], ...]) -> None:
    for old_value, new_value in renames:
        bind.execute(
            sa.text("UPDATE accounts SET status = :new_value WHERE status = :old_value"),
            {"old_value": old_value, "new_value": new_value},
        )


def upgrade() -> None:
    bind = op.get_bind()
    if bind.dialect.name != "sqlite":
        return
    _apply_status_renames(bind, _ACCOUNT_STATUS_RENAMES)


def downgrade() -> None:
    bind = op.get_bind()
    if bind.dialect.name != "sqlite":
        return
    reverse_renames = tuple((new_value, old_value) for old_value, new_value in _ACCOUNT_STATUS_RENAMES)
    _apply_status_renames(bind, reverse_renames)



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import cast

import pytest

from app.modules.firewall.repository import FirewallRepositoryConflictError
from app.modules.firewall.service import (
    FirewallIpAlreadyExistsError,
    FirewallRepositoryPort,
    FirewallService,
    FirewallValidationError,
    normalize_ip_address,
)

pytestmark = pytest.mark.unit


@dataclass(slots=True)
class _Entry:
    ip_address: str
    created_at: datetime


class _Repo:
    def __init__(self) -> None:
        self._entries: dict[str, _Entry] = {}

    async def list_entries(self) -> Sequence[_Entry]:
        return sorted(self._entries.values(), key=lambda entry: (entry.created_at, entry.ip_address))

    async def list_ip_addresses(self) -> set[str]:
        return set(self._entries.keys())

    async def exists(self, ip_address: str) -> bool:
        return ip_address in self._entries

    async def add(self, ip_address: str) -> _Entry:
        entry = _Entry(ip_address=ip_address, created_at=datetime.now(UTC))
        self._entries[ip_address] = entry
        return entry

    async def delete(self, ip_address: str) -> bool:
        return self._entries.pop(ip_address, None) is not None


def test_normalize_ip_address_rejects_invalid_value():
    with pytest.raises(FirewallValidationError):
        normalize_ip_address("invalid")


def test_normalize_ip_address_normalizes_ipv6():
    assert normalize_ip_address("2001:0db8:0000:0000:0000:ff00:0042:8329") == "2001:db8::ff00:42:8329"


@pytest.mark.asyncio
async def test_add_ip_rejects_duplicates():
    service = FirewallService(cast(FirewallRepositoryPort, _Repo()))
    await service.add_ip("127.0.0.1")
    with pytest.raises(FirewallIpAlreadyExistsError):
        await service.add_ip("127.0.0.1")


@pytest.mark.asyncio
async def test_add_ip_maps_repository_conflict_to_exists_error():
    class _ConflictRepo(_Repo):
        async def exists(self, ip_address: str) -> bool:
            return False

        async def add(self, ip_address: str) -> _Entry:
            raise FirewallRepositoryConflictError("duplicate")

    service = FirewallService(cast(FirewallRepositoryPort, _ConflictRepo()))
    with pytest.raises(FirewallIpAlreadyExistsError):
        await service.add_ip("127.0.0.1")


@pytest.mark.asyncio
async def test_is_ip_allowed_follows_allowlist_mode():
    service = FirewallService(cast(FirewallRepositoryPort, _Repo()))

    assert await service.is_ip_allowed("192.168.0.1") is True

    await service.add_ip("127.0.0.1")

    assert await service.is_ip_allowed("127.0.0.1") is True
    assert await service.is_ip_allowed("192.168.0.1") is False
    assert await service.is_ip_allowed("invalid-ip") is False



"""merge request-log transport and feature heads

Revision ID: 20260312_020000_merge_request_logs_transport_and_feature_heads
Revises: 20260310_000000_add_request_logs_transport, 20260312_010000_merge_additional_usage_and_sticky_session_heads
Create Date: 2026-03-12
"""

from __future__ import annotations

# revision identifiers, used by Alembic.
revision = "20260312_020000_merge_request_logs_transport_and_feature_heads"
down_revision = (
    "20260310_000000_add_request_logs_transport",
    "20260312_010000_merge_additional_usage_and_sticky_session_heads",
)
branch_labels = None
depends_on = None


def upgrade() -> None:
    return


def downgrade() -> None:
    return



from __future__ import annotations

import asyncio
import time

_draining: bool = False
_bridge_drain_active: bool = False
_in_flight: int = 0


def reset() -> None:
    global _draining, _bridge_drain_active, _in_flight
    _draining = False
    _bridge_drain_active = False
    _in_flight = 0


def set_draining(val: bool = True) -> None:
    global _draining
    _draining = val


def is_draining() -> bool:
    return _draining


def set_bridge_drain_active(val: bool = True) -> None:
    global _bridge_drain_active
    _bridge_drain_active = val


def is_bridge_drain_active() -> bool:
    return _bridge_drain_active


def increment_in_flight() -> None:
    global _in_flight
    _in_flight += 1


def decrement_in_flight() -> None:
    global _in_flight
    _in_flight = max(0, _in_flight - 1)


def get_in_flight() -> int:
    return _in_flight


async def wait_for_in_flight_drain(timeout_seconds: float, poll_interval_seconds: float = 0.1) -> bool:
    deadline = time.monotonic() + timeout_seconds
    while get_in_flight() > 0 and time.monotonic() < deadline:
        await asyncio.sleep(poll_interval_seconds)
    return get_in_flight() == 0



from app.core.balancer.logic import (
    HEALTH_TIER_DRAINING,
    HEALTH_TIER_HEALTHY,
    HEALTH_TIER_PROBING,
    PERMANENT_FAILURE_CODES,
    AccountState,
    FailoverAction,
    RoutingStrategy,
    SelectionResult,
    evaluate_health_tier,
    failover_decision,
    handle_permanent_failure,
    handle_quota_exceeded,
    handle_rate_limit,
    select_account,
)

__all__ = [
    "PERMANENT_FAILURE_CODES",
    "HEALTH_TIER_DRAINING",
    "HEALTH_TIER_HEALTHY",
    "HEALTH_TIER_PROBING",
    "AccountState",
    "FailoverAction",
    "RoutingStrategy",
    "SelectionResult",
    "evaluate_health_tier",
    "failover_decision",
    "handle_permanent_failure",
    "handle_quota_exceeded",
    "handle_rate_limit",
    "select_account",
]



from __future__ import annotations

import pytest

from app.core.utils.request_id import ensure_request_id, get_request_id, reset_request_id, set_request_id

pytestmark = pytest.mark.unit


def test_ensure_request_id_uses_explicit_value() -> None:
    token = set_request_id(None)
    try:
        assert ensure_request_id("req-explicit") == "req-explicit"
        assert get_request_id() == "req-explicit"
    finally:
        reset_request_id(token)


def test_ensure_request_id_generates_and_reuses_request_id() -> None:
    token = set_request_id(None)
    try:
        request_id = ensure_request_id()
        assert request_id == ensure_request_id()
        assert get_request_id() == request_id
        assert request_id
    finally:
        reset_request_id(token)



from __future__ import annotations

import base64
import json

import pytest
from sqlalchemy import select

import app.modules.proxy.service as proxy_module
from app.core.auth import generate_unique_account_id
from app.core.clients.proxy import ProxyResponseError
from app.db.models import Account, AccountStatus, RequestLog
from app.db.session import SessionLocal

pytestmark = pytest.mark.integration


@pytest.fixture(autouse=True)
async def _force_usage_weighted_routing(async_client) -> None:
    current = await async_client.get("/api/settings")
    assert current.status_code == 200
    payload = current.json()
    payload["routingStrategy"] = "usage_weighted"
    response = await async_client.put("/api/settings", json=payload)
    assert response.status_code == 200


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


def _sse_event(payload: dict) -> str:
    return f"data: {json.dumps(payload)}\n\n"


def _extract_first_event(lines: list[str]) -> dict:
    for line in lines:
        if line.startswith("data: "):
            return json.loads(line[6:])
    raise AssertionError("No SSE data event found")


async def _import_account(async_client, account_id: str, email: str) -> str:
    auth_json = _make_auth_json(account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    return generate_unique_account_id(account_id, email)


@pytest.mark.asyncio
async def test_proxy_compact_not_implemented(async_client, monkeypatch):
    await _import_account(async_client, "acc_compact_ni", "ni@example.com")

    async def fake_compact(*_args, **_kwargs):
        raise NotImplementedError

    monkeypatch.setattr(proxy_module, "core_compact_responses", fake_compact)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": []}
    response = await async_client.post("/backend-api/codex/responses/compact", json=payload)
    assert response.status_code == 501
    assert response.json()["error"]["code"] == "not_implemented"


@pytest.mark.asyncio
async def test_proxy_compact_upstream_error_propagates(async_client, monkeypatch):
    await _import_account(async_client, "acc_compact_err", "err@example.com")

    async def fake_compact(*_args, **_kwargs):
        raise ProxyResponseError(502, {"error": {"code": "upstream_error", "message": "boom"}})

    monkeypatch.setattr(proxy_module, "core_compact_responses", fake_compact)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": []}
    response = await async_client.post("/backend-api/codex/responses/compact", json=payload)
    assert response.status_code == 502
    assert response.json()["error"]["code"] == "upstream_error"


@pytest.mark.asyncio
async def test_proxy_stream_records_cached_and_reasoning_tokens(async_client, monkeypatch):
    expected_account_id = await _import_account(async_client, "acc_usage", "usage@example.com")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        usage = {
            "input_tokens": 10,
            "output_tokens": 5,
            "input_tokens_details": {"cached_tokens": 3},
            "output_tokens_details": {"reasoning_tokens": 2},
        }
        event = {"type": "response.completed", "response": {"id": "resp_1", "usage": usage}}
        yield _sse_event(event)

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True}
    request_id = "req_usage_123"
    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json=payload,
        headers={"x-request-id": request_id},
    ) as resp:
        assert resp.status_code == 200
        lines = [line async for line in resp.aiter_lines() if line]

    event = _extract_first_event(lines)
    assert event["type"] == "response.completed"

    async with SessionLocal() as session:
        result = await session.execute(
            select(RequestLog)
            .where(RequestLog.account_id == expected_account_id)
            .order_by(RequestLog.requested_at.desc())
        )
        log = result.scalars().first()
        assert log is not None
        assert log.request_id == request_id
        assert log.input_tokens == 10
        assert log.output_tokens == 5
        assert log.cached_input_tokens == 3
        assert log.reasoning_tokens == 2
        assert log.status == "success"


@pytest.mark.asyncio
async def test_proxy_stream_retries_rate_limit_then_success(async_client, monkeypatch):
    expected_account_id_1 = await _import_account(async_client, "acc_1", "one@example.com")
    expected_account_id_2 = await _import_account(async_client, "acc_2", "two@example.com")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        if account_id == "acc_1":
            event = {
                "type": "response.failed",
                "response": {"error": {"code": "rate_limit_exceeded", "message": "slow down"}},
            }
            yield _sse_event(event)
            return
        event = {
            "type": "response.completed",
            "response": {"id": "resp_2", "usage": {"input_tokens": 1, "output_tokens": 1}},
        }
        yield _sse_event(event)

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True}
    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json=payload,
    ) as resp:
        assert resp.status_code == 200
        lines = [line async for line in resp.aiter_lines() if line]

    event = _extract_first_event(lines)
    assert event["type"] == "response.completed"

    async with SessionLocal() as session:
        result = await session.execute(select(RequestLog).order_by(RequestLog.requested_at.desc()))
        logs = list(result.scalars().all())
        assert len(logs) == 2
        by_account = {log.account_id: log for log in logs}
        assert by_account[expected_account_id_1].status == "error"
        assert by_account[expected_account_id_1].error_code == "rate_limit_exceeded"
        assert by_account[expected_account_id_1].error_message == "slow down"
        assert by_account[expected_account_id_2].status == "success"

    async with SessionLocal() as session:
        acc1 = await session.get(Account, expected_account_id_1)
        acc2 = await session.get(Account, expected_account_id_2)
        assert acc1 is not None
        assert acc2 is not None
        assert acc1.status == AccountStatus.RATE_LIMITED
        assert acc2.status == AccountStatus.ACTIVE


@pytest.mark.asyncio
async def test_proxy_stream_does_not_retry_stream_idle_timeout(async_client, monkeypatch):
    await _import_account(async_client, "acc_idle_1", "idle-one@example.com")
    await _import_account(async_client, "acc_idle_2", "idle-two@example.com")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        event = {
            "type": "response.failed",
            "response": {"error": {"code": "stream_idle_timeout", "message": "idle"}},
        }
        yield _sse_event(event)

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True}
    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json=payload,
    ) as resp:
        assert resp.status_code == 200
        lines = [line async for line in resp.aiter_lines() if line]

    event = _extract_first_event(lines)
    assert event["type"] == "response.failed"
    assert event["response"]["error"]["code"] == "stream_idle_timeout"

    async with SessionLocal() as session:
        result = await session.execute(select(RequestLog).order_by(RequestLog.requested_at.desc()))
        logs = list(result.scalars().all())
        assert len(logs) == 1
        assert logs[0].error_code == "stream_idle_timeout"


@pytest.mark.asyncio
async def test_proxy_stream_drops_forwarded_headers(async_client, monkeypatch):
    await _import_account(async_client, "acc_headers", "headers@example.com")
    captured_headers: dict[str, str] = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        captured_headers.update(headers)
        event = {
            "type": "response.completed",
            "response": {"id": "resp_headers", "usage": {"input_tokens": 1, "output_tokens": 1}},
        }
        yield _sse_event(event)

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True}
    request_headers = {
        "x-forwarded-for": "1.2.3.4",
        "x-forwarded-proto": "https",
        "x-real-ip": "1.2.3.4",
        "forwarded": "for=1.2.3.4;proto=https",
        "cf-connecting-ip": "1.2.3.4",
        "cf-ray": "ray123",
        "true-client-ip": "1.2.3.4",
        "user-agent": "codex-test",
    }
    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json=payload,
        headers=request_headers,
    ) as resp:
        assert resp.status_code == 200
        _ = [line async for line in resp.aiter_lines() if line]

    normalized = {key.lower() for key in captured_headers}
    assert "x-forwarded-for" not in normalized
    assert "x-forwarded-proto" not in normalized
    assert "x-real-ip" not in normalized
    assert "forwarded" not in normalized
    assert "cf-connecting-ip" not in normalized
    assert "cf-ray" not in normalized
    assert "true-client-ip" not in normalized
    assert "user-agent" in normalized


@pytest.mark.asyncio
async def test_proxy_stream_usage_limit_returns_http_error(async_client, monkeypatch):
    expected_account_id = await _import_account(async_client, "acc_limit", "limit@example.com")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        raise ProxyResponseError(
            429,
            {
                "error": {
                    "type": "usage_limit_reached",
                    "message": "The usage limit has been reached",
                    "plan_type": "plus",
                    "resets_at": 1767612327,
                }
            },
        )
        if False:
            yield ""

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True}
    response = await async_client.post("/backend-api/codex/responses", json=payload)
    assert response.status_code == 429
    error = response.json()["error"]
    assert error["type"] == "usage_limit_reached"
    assert error["plan_type"] == "plus"
    assert error["resets_at"] == 1767612327

    async with SessionLocal() as session:
        acc = await session.get(Account, expected_account_id)
        assert acc is not None
        assert acc.status == AccountStatus.RATE_LIMITED



from __future__ import annotations

import pytest

pytestmark = pytest.mark.integration


@pytest.mark.asyncio
async def test_settings_api_get_and_update(async_client):
    response = await async_client.get("/api/settings")
    assert response.status_code == 200
    payload = response.json()
    assert payload["stickyThreadsEnabled"] is False
    assert payload["upstreamStreamTransport"] == "default"
    assert payload["preferEarlierResetAccounts"] is False
    assert payload["routingStrategy"] == "capacity_weighted"
    assert payload["openaiCacheAffinityMaxAgeSeconds"] == 1800
    assert payload["httpResponsesSessionBridgePromptCacheIdleTtlSeconds"] == 3600
    assert payload["httpResponsesSessionBridgeGatewaySafeMode"] is False
    assert payload["stickyReallocationBudgetThresholdPct"] == 95.0
    assert payload["importWithoutOverwrite"] is True
    assert payload["totpRequiredOnLogin"] is False
    assert payload["totpConfigured"] is False
    assert payload["apiKeyAuthEnabled"] is False

    response = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": True,
            "upstreamStreamTransport": "websocket",
            "preferEarlierResetAccounts": True,
            "routingStrategy": "round_robin",
            "openaiCacheAffinityMaxAgeSeconds": 180,
            "httpResponsesSessionBridgePromptCacheIdleTtlSeconds": 1800,
            "httpResponsesSessionBridgeGatewaySafeMode": True,
            "stickyReallocationBudgetThresholdPct": 90.0,
            "importWithoutOverwrite": False,
            "totpRequiredOnLogin": False,
            "apiKeyAuthEnabled": True,
        },
    )
    assert response.status_code == 200
    updated = response.json()
    assert updated["stickyThreadsEnabled"] is True
    assert updated["upstreamStreamTransport"] == "websocket"
    assert updated["preferEarlierResetAccounts"] is True
    assert updated["routingStrategy"] == "round_robin"
    assert updated["openaiCacheAffinityMaxAgeSeconds"] == 180
    assert updated["httpResponsesSessionBridgePromptCacheIdleTtlSeconds"] == 1800
    assert updated["httpResponsesSessionBridgeGatewaySafeMode"] is True
    assert updated["stickyReallocationBudgetThresholdPct"] == 90.0
    assert updated["importWithoutOverwrite"] is False
    assert updated["totpRequiredOnLogin"] is False
    assert updated["totpConfigured"] is False
    assert updated["apiKeyAuthEnabled"] is True

    response = await async_client.get("/api/settings")
    assert response.status_code == 200
    payload = response.json()
    assert payload["stickyThreadsEnabled"] is True
    assert payload["upstreamStreamTransport"] == "websocket"
    assert payload["preferEarlierResetAccounts"] is True
    assert payload["routingStrategy"] == "round_robin"
    assert payload["openaiCacheAffinityMaxAgeSeconds"] == 180
    assert payload["httpResponsesSessionBridgePromptCacheIdleTtlSeconds"] == 1800
    assert payload["httpResponsesSessionBridgeGatewaySafeMode"] is True
    assert payload["stickyReallocationBudgetThresholdPct"] == 90.0
    assert payload["importWithoutOverwrite"] is False
    assert payload["totpRequiredOnLogin"] is False
    assert payload["totpConfigured"] is False
    assert payload["apiKeyAuthEnabled"] is True



from __future__ import annotations

import uuid
from datetime import datetime
from enum import Enum

from sqlalchemy import (
    BigInteger,
    Boolean,
    CheckConstraint,
    DateTime,
    Float,
    ForeignKey,
    Index,
    Integer,
    LargeBinary,
    String,
    Text,
    UniqueConstraint,
    false,
    func,
    literal_column,
    text,
    true,
)
from sqlalchemy import Enum as SqlEnum
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


class Base(DeclarativeBase):
    pass


def _enum_values(enum_cls: type[Enum]) -> list[str]:
    return [str(member.value) for member in enum_cls]


class AccountStatus(str, Enum):
    ACTIVE = "active"
    RATE_LIMITED = "rate_limited"
    QUOTA_EXCEEDED = "quota_exceeded"
    PAUSED = "paused"
    DEACTIVATED = "deactivated"


class StickySessionKind(str, Enum):
    CODEX_SESSION = "codex_session"
    STICKY_THREAD = "sticky_thread"
    PROMPT_CACHE = "prompt_cache"


class UpstreamProviderKind(str, Enum):
    CHATGPT_WEB = "chatgpt_web"
    OPENAI_PLATFORM = "openai_platform"


class Account(Base):
    __tablename__ = "accounts"

    id: Mapped[str] = mapped_column(String, primary_key=True)
    chatgpt_account_id: Mapped[str | None] = mapped_column(String, nullable=True)
    email: Mapped[str] = mapped_column(String, nullable=False)
    plan_type: Mapped[str] = mapped_column(String, nullable=False)

    access_token_encrypted: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)
    refresh_token_encrypted: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)
    id_token_encrypted: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)

    last_refresh: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)

    status: Mapped[AccountStatus] = mapped_column(
        SqlEnum(
            AccountStatus,
            name="account_status",
            validate_strings=True,
            values_callable=_enum_values,
        ),
        default=AccountStatus.ACTIVE,
        nullable=False,
    )
    deactivation_reason: Mapped[str | None] = mapped_column(Text, nullable=True)
    reset_at: Mapped[int | None] = mapped_column(Integer, nullable=True)

    api_key_assignments: Mapped[list["ApiKeyAccountAssignment"]] = relationship(
        "ApiKeyAccountAssignment",
        back_populates="account",
        cascade="all, delete-orphan",
    )


class UsageHistory(Base):
    __tablename__ = "usage_history"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    account_id: Mapped[str] = mapped_column(String, ForeignKey("accounts.id", ondelete="CASCADE"), nullable=False)
    recorded_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    window: Mapped[str | None] = mapped_column(String, nullable=True)
    used_percent: Mapped[float] = mapped_column(Float, nullable=False)
    input_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    output_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    reset_at: Mapped[int | None] = mapped_column(Integer, nullable=True)
    window_minutes: Mapped[int | None] = mapped_column(Integer, nullable=True)
    credits_has: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    credits_unlimited: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    credits_balance: Mapped[float | None] = mapped_column(Float, nullable=True)


class AdditionalUsageHistory(Base):
    __tablename__ = "additional_usage_history"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    account_id: Mapped[str] = mapped_column(String, ForeignKey("accounts.id", ondelete="CASCADE"), nullable=False)
    quota_key: Mapped[str] = mapped_column(String, nullable=False)
    limit_name: Mapped[str] = mapped_column(String, nullable=False)
    metered_feature: Mapped[str] = mapped_column(String, nullable=False)
    window: Mapped[str] = mapped_column(String, nullable=False)
    used_percent: Mapped[float] = mapped_column(Float, nullable=False)
    reset_at: Mapped[int | None] = mapped_column(Integer, nullable=True)
    window_minutes: Mapped[int | None] = mapped_column(Integer, nullable=True)
    recorded_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)


class RequestLog(Base):
    __tablename__ = "request_logs"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    account_id: Mapped[str | None] = mapped_column(String, ForeignKey("accounts.id", ondelete="CASCADE"), nullable=True)
    provider_kind: Mapped[str | None] = mapped_column(String, nullable=True)
    routing_subject_id: Mapped[str | None] = mapped_column(String, nullable=True)
    api_key_id: Mapped[str | None] = mapped_column(String, nullable=True)
    request_id: Mapped[str] = mapped_column(String, nullable=False)
    requested_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    model: Mapped[str] = mapped_column(String, nullable=False)
    transport: Mapped[str | None] = mapped_column(String, nullable=True)
    route_class: Mapped[str | None] = mapped_column(String, nullable=True)
    upstream_request_id: Mapped[str | None] = mapped_column(String, nullable=True)
    rejection_reason: Mapped[str | None] = mapped_column(String, nullable=True)
    service_tier: Mapped[str | None] = mapped_column(String, nullable=True)
    requested_service_tier: Mapped[str | None] = mapped_column(String, nullable=True)
    actual_service_tier: Mapped[str | None] = mapped_column(String, nullable=True)
    input_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    output_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    cached_input_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    reasoning_tokens: Mapped[int | None] = mapped_column(Integer, nullable=True)
    cost_usd: Mapped[float | None] = mapped_column(Float, nullable=True)
    reasoning_effort: Mapped[str | None] = mapped_column(String, nullable=True)
    latency_ms: Mapped[int | None] = mapped_column(Integer, nullable=True)
    latency_first_token_ms: Mapped[int | None] = mapped_column(Integer, nullable=True)
    status: Mapped[str] = mapped_column(String, nullable=False)
    error_code: Mapped[str | None] = mapped_column(String, nullable=True)
    error_message: Mapped[str | None] = mapped_column(Text, nullable=True)


class AuditLog(Base):
    __tablename__ = "audit_logs"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=func.now(), nullable=False, index=True)
    action: Mapped[str] = mapped_column(String(100), nullable=False, index=True)
    actor_ip: Mapped[str | None] = mapped_column(String(50), nullable=True)
    details: Mapped[str | None] = mapped_column(Text, nullable=True)
    request_id: Mapped[str | None] = mapped_column(String(100), nullable=True)


class SchedulerLeader(Base):
    __tablename__ = "scheduler_leader"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, default=1)
    leader_id: Mapped[str] = mapped_column(String(100), nullable=False)
    acquired_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    expires_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False, index=True)


class StickySession(Base):
    __tablename__ = "sticky_sessions"
    __table_args__ = (
        CheckConstraint("routing_subject_id <> ''", name="ck_sticky_sessions_routing_subject_non_empty"),
        CheckConstraint(
            "account_id IS NULL OR provider_kind = 'chatgpt_web'",
            name="ck_sticky_sessions_account_scope",
        ),
        CheckConstraint(
            "NOT (provider_kind = 'openai_platform' AND kind = 'codex_session')",
            name="ck_sticky_sessions_platform_codex_session",
        ),
    )

    key: Mapped[str] = mapped_column(String, primary_key=True)
    kind: Mapped[StickySessionKind] = mapped_column(
        SqlEnum(
            StickySessionKind,
            name="sticky_session_kind",
            validate_strings=True,
            values_callable=_enum_values,
        ),
        primary_key=True,
        default=StickySessionKind.STICKY_THREAD,
        server_default=text("'sticky_thread'"),
        nullable=False,
    )
    provider_kind: Mapped[str] = mapped_column(
        String,
        primary_key=True,
        default=UpstreamProviderKind.CHATGPT_WEB.value,
        server_default=text("'chatgpt_web'"),
        nullable=False,
    )
    account_id: Mapped[str | None] = mapped_column(String, ForeignKey("accounts.id", ondelete="CASCADE"), nullable=True)
    routing_subject_id: Mapped[str] = mapped_column(String, nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )


class OpenAIPlatformIdentity(Base):
    __tablename__ = "openai_platform_identities"
    __table_args__ = (Index("uq_openai_platform_identities_singleton", "singleton_key", unique=True),)

    id: Mapped[str] = mapped_column(String, primary_key=True)
    singleton_key: Mapped[int] = mapped_column(Integer, nullable=False, default=1, server_default=text("1"))
    label: Mapped[str] = mapped_column(String, nullable=False)
    api_key_encrypted: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)
    organization_id: Mapped[str | None] = mapped_column(String, nullable=True)
    project_id: Mapped[str | None] = mapped_column(String, nullable=True)
    eligible_route_families: Mapped[str] = mapped_column(Text, nullable=False, default="", server_default=text("''"))
    status: Mapped[AccountStatus] = mapped_column(
        SqlEnum(
            AccountStatus,
            name="account_status",
            validate_strings=True,
            values_callable=_enum_values,
        ),
        default=AccountStatus.ACTIVE,
        server_default=text("'active'"),
        nullable=False,
    )
    last_validated_at: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)
    last_auth_failure_reason: Mapped[str | None] = mapped_column(Text, nullable=True)
    deactivation_reason: Mapped[str | None] = mapped_column(Text, nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )


class DashboardSettings(Base):
    __tablename__ = "dashboard_settings"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=False)
    sticky_threads_enabled: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)
    upstream_stream_transport: Mapped[str] = mapped_column(
        String,
        default="default",
        server_default=text("'default'"),
        nullable=False,
    )
    prefer_earlier_reset_accounts: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)
    routing_strategy: Mapped[str] = mapped_column(
        String,
        default="capacity_weighted",
        server_default=text("'capacity_weighted'"),
        nullable=False,
    )
    openai_cache_affinity_max_age_seconds: Mapped[int] = mapped_column(
        Integer,
        default=1800,
        server_default=text("1800"),
        nullable=False,
    )
    import_without_overwrite: Mapped[bool] = mapped_column(
        Boolean,
        default=True,
        server_default=true(),
        nullable=False,
    )
    totp_required_on_login: Mapped[bool] = mapped_column(
        Boolean,
        default=False,
        nullable=False,
    )
    password_hash: Mapped[str | None] = mapped_column(Text, nullable=True)
    api_key_auth_enabled: Mapped[bool] = mapped_column(
        Boolean,
        default=False,
        nullable=False,
    )
    totp_secret_encrypted: Mapped[bytes | None] = mapped_column(LargeBinary, nullable=True)
    totp_last_verified_step: Mapped[int | None] = mapped_column(Integer, nullable=True)
    http_responses_session_bridge_prompt_cache_idle_ttl_seconds: Mapped[int] = mapped_column(
        Integer,
        default=3600,
        server_default=text("3600"),
        nullable=False,
    )
    http_responses_session_bridge_gateway_safe_mode: Mapped[bool] = mapped_column(
        Boolean,
        default=False,
        server_default=false(),
        nullable=False,
    )
    sticky_reallocation_budget_threshold_pct: Mapped[float] = mapped_column(
        Float,
        default=95.0,
        server_default=text("95.0"),
        nullable=False,
    )
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )


class ApiFirewallAllowlist(Base):
    __tablename__ = "api_firewall_allowlist"

    ip_address: Mapped[str] = mapped_column(String, primary_key=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)


class ApiKey(Base):
    __tablename__ = "api_keys"

    id: Mapped[str] = mapped_column(String, primary_key=True)
    name: Mapped[str] = mapped_column(String, nullable=False)
    key_hash: Mapped[str] = mapped_column(String, nullable=False, unique=True)
    key_prefix: Mapped[str] = mapped_column(String, nullable=False)
    allowed_models: Mapped[str | None] = mapped_column(Text, nullable=True)
    enforced_model: Mapped[str | None] = mapped_column(String, nullable=True)
    enforced_reasoning_effort: Mapped[str | None] = mapped_column(String, nullable=True)
    enforced_service_tier: Mapped[str | None] = mapped_column(String, nullable=True)
    account_assignment_scope_enabled: Mapped[bool] = mapped_column(
        Boolean,
        default=False,
        server_default=false(),
        nullable=False,
    )
    expires_at: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)
    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    last_used_at: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)

    limits: Mapped[list["ApiKeyLimit"]] = relationship(
        "ApiKeyLimit",
        back_populates="api_key",
        cascade="all, delete-orphan",
        lazy="selectin",
    )
    account_assignments: Mapped[list["ApiKeyAccountAssignment"]] = relationship(
        "ApiKeyAccountAssignment",
        back_populates="api_key",
        cascade="all, delete-orphan",
        lazy="selectin",
    )


class ApiKeyAccountAssignment(Base):
    __tablename__ = "api_key_accounts"

    api_key_id: Mapped[str] = mapped_column(
        String,
        ForeignKey("api_keys.id", ondelete="CASCADE"),
        primary_key=True,
    )
    account_id: Mapped[str] = mapped_column(
        String,
        ForeignKey("accounts.id", ondelete="CASCADE"),
        primary_key=True,
    )
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)

    api_key: Mapped["ApiKey"] = relationship("ApiKey", back_populates="account_assignments")
    account: Mapped["Account"] = relationship("Account", back_populates="api_key_assignments")


class LimitType(str, Enum):
    TOTAL_TOKENS = "total_tokens"
    INPUT_TOKENS = "input_tokens"
    OUTPUT_TOKENS = "output_tokens"
    COST_USD = "cost_usd"
    CREDITS = "credits"


class LimitWindow(str, Enum):
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    FIVE_HOURS = "5h"
    SEVEN_DAYS = "7d"


class ApiKeyLimit(Base):
    __tablename__ = "api_key_limits"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    api_key_id: Mapped[str] = mapped_column(
        String,
        ForeignKey("api_keys.id", ondelete="CASCADE"),
        nullable=False,
    )
    limit_type: Mapped[LimitType] = mapped_column(
        SqlEnum(
            LimitType,
            name="limit_type",
            validate_strings=True,
            values_callable=_enum_values,
        ),
        nullable=False,
    )
    limit_window: Mapped[LimitWindow] = mapped_column(
        SqlEnum(
            LimitWindow,
            name="limit_window",
            validate_strings=True,
            values_callable=_enum_values,
        ),
        nullable=False,
    )
    max_value: Mapped[int] = mapped_column(BigInteger, nullable=False)
    current_value: Mapped[int] = mapped_column(BigInteger, default=0, server_default=text("0"), nullable=False)
    model_filter: Mapped[str | None] = mapped_column(String, nullable=True)
    reset_at: Mapped[datetime] = mapped_column(DateTime, nullable=False)

    api_key: Mapped["ApiKey"] = relationship("ApiKey", back_populates="limits")


class ApiKeyUsageReservation(Base):
    __tablename__ = "api_key_usage_reservations"

    id: Mapped[str] = mapped_column(String, primary_key=True)
    api_key_id: Mapped[str] = mapped_column(
        String,
        ForeignKey("api_keys.id", ondelete="CASCADE"),
        nullable=False,
    )
    model: Mapped[str] = mapped_column(String, nullable=False)
    status: Mapped[str] = mapped_column(String, nullable=False, default="reserved")
    input_tokens: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    output_tokens: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    cached_input_tokens: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    cost_microdollars: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    items: Mapped[list["ApiKeyUsageReservationItem"]] = relationship(
        "ApiKeyUsageReservationItem",
        back_populates="reservation",
        cascade="all, delete-orphan",
        lazy="selectin",
    )


class ApiKeyUsageReservationItem(Base):
    __tablename__ = "api_key_usage_reservation_items"
    __table_args__ = (UniqueConstraint("reservation_id", "limit_id", name="uq_reservation_limit"),)

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    reservation_id: Mapped[str] = mapped_column(
        String,
        ForeignKey("api_key_usage_reservations.id", ondelete="CASCADE"),
        nullable=False,
    )
    limit_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("api_key_limits.id", ondelete="CASCADE"),
        nullable=False,
    )
    limit_type: Mapped[str] = mapped_column(String, nullable=False)
    reserved_delta: Mapped[int] = mapped_column(BigInteger, nullable=False)
    actual_delta: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    expected_reset_at: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    reservation: Mapped[ApiKeyUsageReservation] = relationship(
        "ApiKeyUsageReservation",
        back_populates="items",
    )
    limit: Mapped[ApiKeyLimit] = relationship("ApiKeyLimit")


class RateLimitAttempt(Base):
    __tablename__ = "rate_limit_attempts"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    key: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    attempted_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=func.now(), nullable=False)
    type: Mapped[str] = mapped_column(String(50), nullable=False)


class CacheInvalidation(Base):
    __tablename__ = "cache_invalidation"

    namespace: Mapped[str] = mapped_column(String(50), primary_key=True)
    version: Mapped[int] = mapped_column(Integer, nullable=False, server_default="0")


class BridgeRingMember(Base):
    __tablename__ = "bridge_ring_members"

    id: Mapped[str] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()))
    instance_id: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    registered_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False, default=func.now())
    last_heartbeat_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False, default=func.now())
    metadata_json: Mapped[str | None] = mapped_column(Text, nullable=True)


_PRIMARY_WINDOW_INDEX_EXPR = func.coalesce(UsageHistory.window, literal_column("'primary'"))

Index("idx_usage_recorded_at", UsageHistory.recorded_at)
Index("idx_usage_account_time", UsageHistory.account_id, UsageHistory.recorded_at)
Index(
    "idx_usage_window_account_time",
    _PRIMARY_WINDOW_INDEX_EXPR,
    UsageHistory.account_id,
    UsageHistory.recorded_at,
)
Index(
    "idx_usage_window_account_latest",
    _PRIMARY_WINDOW_INDEX_EXPR,
    UsageHistory.account_id,
    UsageHistory.recorded_at.desc(),
    UsageHistory.id.desc(),
)
Index("idx_accounts_email", Account.email)
Index("idx_openai_platform_identities_status", OpenAIPlatformIdentity.status)
Index("idx_openai_platform_identities_label", OpenAIPlatformIdentity.label)
Index("idx_api_keys_name", ApiKey.name)
Index("idx_logs_account_time", RequestLog.account_id, RequestLog.requested_at)
Index("idx_logs_provider_time", RequestLog.provider_kind, RequestLog.requested_at)
Index("idx_logs_routing_subject_time", RequestLog.routing_subject_id, RequestLog.requested_at)
Index("idx_logs_requested_at", RequestLog.requested_at)
Index("idx_logs_requested_at_id", RequestLog.requested_at.desc(), RequestLog.id.desc())
Index(
    "idx_logs_requested_at_model_tier",
    RequestLog.requested_at.desc(),
    RequestLog.model,
    RequestLog.service_tier,
)
Index(
    "idx_logs_model_effort_time",
    RequestLog.model,
    RequestLog.reasoning_effort,
    RequestLog.requested_at.desc(),
    RequestLog.id.desc(),
)
Index(
    "idx_logs_status_error_time",
    RequestLog.status,
    RequestLog.error_code,
    RequestLog.requested_at.desc(),
    RequestLog.id.desc(),
)
Index("idx_sticky_account", StickySession.account_id)
Index(
    "idx_sticky_provider_routing_kind",
    StickySession.provider_kind,
    StickySession.routing_subject_id,
    StickySession.kind,
)
Index("idx_sticky_kind_updated_at", StickySession.kind, StickySession.updated_at.desc())
Index("idx_api_keys_hash", ApiKey.key_hash)
Index("idx_api_key_accounts_account_id", ApiKeyAccountAssignment.account_id)
Index("idx_api_key_limits_key_id", ApiKeyLimit.api_key_id)
Index("idx_api_key_usage_reservations_key_id", ApiKeyUsageReservation.api_key_id)
Index("idx_api_key_usage_reservations_status", ApiKeyUsageReservation.status)
Index("idx_api_key_usage_res_items_reservation_id", ApiKeyUsageReservationItem.reservation_id)
Index("ix_additional_usage_history_account_id", AdditionalUsageHistory.account_id)
Index("ix_additional_usage_history_recorded_at", AdditionalUsageHistory.recorded_at)
Index(
    "ix_rate_limit_attempts_type_key_attempted_at",
    RateLimitAttempt.type,
    RateLimitAttempt.key,
    RateLimitAttempt.attempted_at,
)
Index(
    "ix_additional_usage_history_composite",
    AdditionalUsageHistory.account_id,
    AdditionalUsageHistory.quota_key,
    AdditionalUsageHistory.window,
    AdditionalUsageHistory.recorded_at,
)
Index(
    "ix_additional_usage_quota_window",
    AdditionalUsageHistory.quota_key,
    AdditionalUsageHistory.window,
    AdditionalUsageHistory.account_id,
    AdditionalUsageHistory.recorded_at,
)



from __future__ import annotations

import json
from collections.abc import Mapping

from app.core.errors import ResponseFailedEvent
from app.core.types import JsonValue
from app.core.utils.json_guards import is_json_dict

type JsonPayload = Mapping[str, JsonValue] | ResponseFailedEvent


def format_sse_event(payload: JsonPayload) -> str:
    data = json.dumps(payload, ensure_ascii=True, separators=(",", ":"))
    event_type = payload.get("type")
    if isinstance(event_type, str) and event_type:
        return f"event: {event_type}\ndata: {data}\n\n"
    return f"data: {data}\n\n"


def format_sse_data(payload: Mapping[str, JsonValue]) -> str:
    data = json.dumps(payload, ensure_ascii=True, separators=(",", ":"))
    return f"data: {data}\n\n"


def parse_sse_data_json(event_block: str) -> dict[str, JsonValue] | None:
    data = extract_sse_data(event_block)
    if data is None:
        return None
    try:
        payload = json.loads(data)
    except json.JSONDecodeError:
        return None
    if is_json_dict(payload):
        return payload
    return None


def extract_sse_data(event_block: str) -> str | None:
    data_lines = _extract_sse_data_lines(event_block)
    if data_lines is None:
        return None
    data = "\n".join(data_lines)
    if not data.strip():
        return None
    if data.strip() == "[DONE]":
        return None
    return data


def _extract_sse_data_lines(event_block: str) -> list[str] | None:
    data_lines: list[str] = []
    for raw_line in event_block.splitlines():
        if not raw_line:
            continue
        if raw_line.startswith(":"):
            continue

        field, value = _parse_sse_field(raw_line)
        if field == "data":
            data_lines.append(value)

    if not data_lines:
        return None
    return data_lines


def _parse_sse_field(line: str) -> tuple[str, str]:
    if ":" not in line:
        return line, ""
    field, value = line.split(":", 1)
    if value.startswith(" "):
        value = value[1:]
    return field, value



from __future__ import annotations

from typing import cast as typing_cast

from app.core.usage.logs import RequestLogLike, cached_input_tokens_from_log, cost_from_log, total_tokens_from_log
from app.db.models import RequestLog
from app.modules.request_logs.schemas import RequestLogEntry

RATE_LIMIT_CODES = {"rate_limit_exceeded", "usage_limit_reached"}
QUOTA_CODES = {"insufficient_quota", "usage_not_included", "quota_exceeded"}


def normalize_log_status(status: str, error_code: str | None) -> str:
    if status == "success":
        return "ok"
    if error_code in RATE_LIMIT_CODES:
        return "rate_limit"
    if error_code in QUOTA_CODES:
        return "quota"
    return "error"


def log_status(log: RequestLog) -> str:
    return normalize_log_status(log.status, log.error_code)


def to_request_log_entry(log: RequestLog, *, api_key_name: str | None = None) -> RequestLogEntry:
    log_like = typing_cast(RequestLogLike, log)
    return RequestLogEntry(
        requested_at=log.requested_at,
        account_id=log.account_id,
        provider_kind=log.provider_kind,
        routing_subject_id=log.routing_subject_id,
        api_key_name=api_key_name,
        request_id=log.request_id,
        model=log.model,
        transport=log.transport,
        route_class=log.route_class,
        upstream_request_id=log.upstream_request_id,
        rejection_reason=log.rejection_reason,
        service_tier=log.service_tier,
        requested_service_tier=log.requested_service_tier,
        actual_service_tier=log.actual_service_tier,
        reasoning_effort=log.reasoning_effort,
        status=log_status(log),
        error_code=log.error_code,
        error_message=log.error_message,
        tokens=total_tokens_from_log(log_like),
        cached_input_tokens=cached_input_tokens_from_log(log_like),
        cost_usd=cost_from_log(log_like, precision=6),
        latency_ms=log.latency_ms,
        latency_first_token_ms=log.latency_first_token_ms,
    )



from __future__ import annotations

from fastapi import APIRouter, Depends, Query

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.openai.model_registry import get_model_registry, is_public_model
from app.dependencies import DashboardContext, get_dashboard_context
from app.modules.dashboard.schemas import DashboardOverviewResponse, DashboardOverviewTimeframeKey

router = APIRouter(
    prefix="/api",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("/dashboard/overview", response_model=DashboardOverviewResponse)
async def get_overview(
    timeframe: DashboardOverviewTimeframeKey = Query("7d"),
    context: DashboardContext = Depends(get_dashboard_context),
) -> DashboardOverviewResponse:
    return await context.service.get_overview(timeframe)


@router.get("/models")
async def list_models() -> dict:
    registry = get_model_registry()
    models_by_slug = registry.get_models_with_fallback()
    if not models_by_slug:
        return {"models": []}
    models = [
        {"id": slug, "name": model.display_name or slug}
        for slug, model in models_by_slug.items()
        if is_public_model(model, None)
    ]
    return {"models": models}



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from fnmatch import fnmatchcase
from typing import Iterable, Mapping

from app.core.openai.models import ResponseUsage
from app.core.usage.types import UsageCostByModel, UsageCostSummary


@dataclass(frozen=True)
class ModelPrice:
    input_per_1m: float
    output_per_1m: float
    cached_input_per_1m: float | None = None
    priority_multiplier: float | None = None
    priority_input_per_1m: float | None = None
    priority_output_per_1m: float | None = None
    priority_cached_input_per_1m: float | None = None
    flex_input_per_1m: float | None = None
    flex_output_per_1m: float | None = None
    flex_cached_input_per_1m: float | None = None
    long_context_threshold_tokens: float | None = None
    long_context_input_per_1m: float | None = None
    long_context_output_per_1m: float | None = None
    long_context_cached_input_per_1m: float | None = None


@dataclass(frozen=True)
class UsageTokens:
    input_tokens: float
    output_tokens: float
    cached_input_tokens: float = 0.0


@dataclass(frozen=True)
class CostItem:
    model: str
    usage: UsageTokens
    service_tier: str | None = None


def _as_number(value: object) -> float | None:
    if isinstance(value, (int, float)):
        return float(value)
    return None


def _normalize_usage(usage: UsageTokens | ResponseUsage | None) -> UsageTokens | None:
    if isinstance(usage, UsageTokens):
        return usage
    if not usage:
        return None
    input_tokens = _as_number(usage.input_tokens)
    output_tokens = _as_number(usage.output_tokens)
    if output_tokens is None and usage.output_tokens_details is not None:
        output_tokens = _as_number(usage.output_tokens_details.reasoning_tokens)
    if input_tokens is None or output_tokens is None:
        return None
    cached_tokens = 0.0
    if usage.input_tokens_details is not None:
        cached_tokens = _as_number(usage.input_tokens_details.cached_tokens) or 0.0
    cached_tokens = max(0.0, min(cached_tokens, input_tokens))
    return UsageTokens(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        cached_input_tokens=cached_tokens,
    )


DEFAULT_PRICING_MODELS: dict[str, ModelPrice] = {
    "gpt-5.4": ModelPrice(
        input_per_1m=2.5,
        cached_input_per_1m=0.25,
        output_per_1m=15.0,
        priority_input_per_1m=5.0,
        priority_cached_input_per_1m=0.5,
        priority_output_per_1m=30.0,
        flex_input_per_1m=1.25,
        flex_cached_input_per_1m=0.125,
        flex_output_per_1m=7.5,
        long_context_threshold_tokens=272_000,
        long_context_input_per_1m=5.0,
        long_context_cached_input_per_1m=0.5,
        long_context_output_per_1m=22.5,
    ),
    "gpt-5.4-mini": ModelPrice(
        input_per_1m=0.75,
        cached_input_per_1m=0.075,
        output_per_1m=4.5,
        flex_input_per_1m=0.375,
        flex_cached_input_per_1m=0.0375,
        flex_output_per_1m=2.25,
    ),
    "gpt-5.4-nano": ModelPrice(
        input_per_1m=0.20,
        cached_input_per_1m=0.02,
        output_per_1m=1.25,
        flex_input_per_1m=0.10,
        flex_cached_input_per_1m=0.01,
        flex_output_per_1m=0.625,
    ),
    "gpt-5.4-pro": ModelPrice(
        input_per_1m=30.0,
        output_per_1m=180.0,
        flex_input_per_1m=15.0,
        flex_output_per_1m=90.0,
        long_context_threshold_tokens=272_000,
        long_context_input_per_1m=60.0,
        long_context_output_per_1m=270.0,
    ),
    "gpt-5.3-codex": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
        priority_input_per_1m=3.5,
        priority_cached_input_per_1m=0.35,
        priority_output_per_1m=28.0,
    ),
    "gpt-5.3": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
    ),
    "gpt-5.3-chat-latest": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
    ),
    "gpt-5.2": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
        priority_multiplier=2.0,
        flex_input_per_1m=0.875,
        flex_cached_input_per_1m=0.0875,
        flex_output_per_1m=7.0,
    ),
    "gpt-5.2-chat-latest": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
    ),
    "gpt-5.1": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
        priority_multiplier=2.0,
        flex_input_per_1m=0.625,
        flex_cached_input_per_1m=0.0625,
        flex_output_per_1m=5.0,
    ),
    "gpt-5.1-chat-latest": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
    ),
    "gpt-5": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
        priority_multiplier=2.0,
        flex_input_per_1m=0.625,
        flex_cached_input_per_1m=0.0625,
        flex_output_per_1m=5.0,
    ),
    "gpt-5-chat-latest": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
    ),
    "gpt-5.2-codex": ModelPrice(
        input_per_1m=1.75,
        cached_input_per_1m=0.175,
        output_per_1m=14.0,
        priority_input_per_1m=3.5,
        priority_cached_input_per_1m=0.35,
        priority_output_per_1m=28.0,
    ),
    "gpt-5.1-codex-max": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
        priority_input_per_1m=2.5,
        priority_cached_input_per_1m=0.25,
        priority_output_per_1m=20.0,
    ),
    "gpt-5.1-codex-mini": ModelPrice(
        input_per_1m=0.25,
        cached_input_per_1m=0.025,
        output_per_1m=2.0,
    ),
    "gpt-5.1-codex": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
        priority_input_per_1m=2.5,
        priority_cached_input_per_1m=0.25,
        priority_output_per_1m=20.0,
    ),
    "gpt-5-codex": ModelPrice(
        input_per_1m=1.25,
        cached_input_per_1m=0.125,
        output_per_1m=10.0,
        priority_input_per_1m=2.5,
        priority_cached_input_per_1m=0.25,
        priority_output_per_1m=20.0,
    ),
}

DEFAULT_MODEL_ALIASES: dict[str, str] = {
    "gpt-5.4-pro*": "gpt-5.4-pro",
    "gpt-5.4-mini*": "gpt-5.4-mini",
    "gpt-5.4-nano*": "gpt-5.4-nano",
    "gpt-5.4*": "gpt-5.4",
    "gpt-5.3-codex*": "gpt-5.3-codex",
    "gpt-5.3-chat-latest*": "gpt-5.3-chat-latest",
    "gpt-5.2-codex*": "gpt-5.2-codex",
    "gpt-5.2-chat-latest*": "gpt-5.2-chat-latest",
    "gpt-5.3*": "gpt-5.3",
    "gpt-5.1-chat-latest*": "gpt-5.1-chat-latest",
    "gpt-5.2*": "gpt-5.2",
    "gpt-5-chat-latest*": "gpt-5-chat-latest",
    "gpt-5.1*": "gpt-5.1",
    "gpt-5*": "gpt-5",
    "gpt-5.1-codex-max*": "gpt-5.1-codex-max",
    "gpt-5.1-codex-mini*": "gpt-5.1-codex-mini",
    "gpt-5.1-codex*": "gpt-5.1-codex",
    "gpt-5-codex*": "gpt-5-codex",
}


def resolve_model_alias(model: str, aliases: Mapping[str, str]) -> str | None:
    if not model:
        return None
    normalized = model.lower()
    matched: list[tuple[int, str]] = []
    for pattern, target in aliases.items():
        if fnmatchcase(normalized, pattern.lower()):
            matched.append((len(pattern), target))
    if not matched:
        return None
    return max(matched, key=lambda item: item[0])[1]


def get_pricing_for_model(
    model: str,
    pricing: Mapping[str, ModelPrice] | None = None,
    aliases: Mapping[str, str] | None = None,
) -> tuple[str, ModelPrice] | None:
    if not model:
        return None
    pricing = pricing or DEFAULT_PRICING_MODELS
    aliases = aliases or DEFAULT_MODEL_ALIASES

    normalized = model.lower()
    for key, value in pricing.items():
        if key.lower() == normalized:
            return key, value

    alias = resolve_model_alias(normalized, aliases)
    if not alias:
        return None
    for key, value in pricing.items():
        if key.lower() == alias.lower():
            return key, value
    return None


def _uses_priority_tier(service_tier: str | None) -> bool:
    normalized = _normalize_service_tier(service_tier)
    if normalized is None:
        return False
    return normalized in {"priority", "fast"}


def _uses_flex_tier(service_tier: str | None) -> bool:
    normalized = _normalize_service_tier(service_tier)
    if normalized is None:
        return False
    return normalized == "flex"


def _normalize_service_tier(service_tier: str | None) -> str | None:
    if service_tier is None:
        return None
    stripped = service_tier.strip().lower()
    return stripped or None


def _effective_rates(
    usage: UsageTokens,
    price: ModelPrice,
    *,
    service_tier: str | None,
) -> tuple[float, float, float]:
    is_long_context = (
        price.long_context_threshold_tokens is not None
        and usage.input_tokens > price.long_context_threshold_tokens
        and price.long_context_input_per_1m is not None
        and price.long_context_output_per_1m is not None
    )
    input_rate = price.input_per_1m
    cached_rate = price.cached_input_per_1m if price.cached_input_per_1m is not None else input_rate
    output_rate = price.output_per_1m

    if _uses_priority_tier(service_tier):
        if price.priority_input_per_1m is not None and price.priority_output_per_1m is not None:
            priority_cached = (
                price.priority_cached_input_per_1m
                if price.priority_cached_input_per_1m is not None
                else price.priority_input_per_1m
            )
            return price.priority_input_per_1m, priority_cached, price.priority_output_per_1m
        if price.priority_multiplier is not None:
            input_rate *= price.priority_multiplier
            cached_rate *= price.priority_multiplier
            output_rate *= price.priority_multiplier
            return input_rate, cached_rate, output_rate

    if _uses_flex_tier(service_tier) and price.flex_input_per_1m is not None and price.flex_output_per_1m is not None:
        input_rate = price.flex_input_per_1m
        cached_rate = price.flex_cached_input_per_1m if price.flex_cached_input_per_1m is not None else input_rate
        output_rate = price.flex_output_per_1m
        if is_long_context:
            input_rate *= 2.0
            cached_rate *= 2.0
            output_rate *= 1.5
        return input_rate, cached_rate, output_rate

    if is_long_context:
        assert price.long_context_input_per_1m is not None
        assert price.long_context_output_per_1m is not None
        input_rate = price.long_context_input_per_1m
        cached_rate = (
            price.long_context_cached_input_per_1m if price.long_context_cached_input_per_1m is not None else input_rate
        )
        output_rate = price.long_context_output_per_1m

    return input_rate, cached_rate, output_rate


def calculate_cost_from_usage(
    usage: UsageTokens | ResponseUsage | None,
    price: ModelPrice,
    *,
    service_tier: str | None = None,
) -> float | None:
    normalized = _normalize_usage(usage)
    if not normalized:
        return None
    billable_input = normalized.input_tokens - normalized.cached_input_tokens

    input_rate, cached_rate, output_rate = _effective_rates(
        normalized,
        price,
        service_tier=service_tier,
    )

    return (
        (billable_input / 1_000_000) * input_rate
        + (normalized.cached_input_tokens / 1_000_000) * cached_rate
        + (normalized.output_tokens / 1_000_000) * output_rate
    )


def calculate_costs(
    items: Iterable[CostItem],
    pricing: Mapping[str, ModelPrice] | None = None,
    aliases: Mapping[str, str] | None = None,
) -> UsageCostSummary:
    pricing = pricing or DEFAULT_PRICING_MODELS
    aliases = aliases or DEFAULT_MODEL_ALIASES

    totals: dict[str, float] = defaultdict(float)
    total_usd = 0.0

    for item in items:
        model = item.model
        usage = item.usage
        resolved = get_pricing_for_model(model, pricing, aliases)
        if not resolved:
            continue
        canonical, price = resolved
        cost = calculate_cost_from_usage(usage, price, service_tier=item.service_tier)
        if cost is None:
            continue
        totals[canonical] += cost
        total_usd += cost

    by_model = [UsageCostByModel(model=model, usd=round(value, 6)) for model, value in sorted(totals.items())]
    return UsageCostSummary(
        currency="USD",
        total_usd_7d=round(total_usd, 6),
        by_model=by_model,
    )



from __future__ import annotations

from dataclasses import dataclass

import bcrypt
import pytest

from app.modules.dashboard_auth.service import (
    DashboardAuthService,
    DashboardSessionStore,
    InvalidCredentialsError,
    PasswordAlreadyConfiguredError,
    PasswordNotConfiguredError,
)

pytestmark = pytest.mark.unit


@dataclass(slots=True)
class _FakeSettings:
    password_hash: str | None = None
    totp_required_on_login: bool = False
    totp_secret_encrypted: bytes | None = None
    totp_last_verified_step: int | None = None


class _FakeRepository:
    def __init__(self) -> None:
        self.settings = _FakeSettings()

    async def get_settings(self) -> _FakeSettings:
        return self.settings

    async def get_password_hash(self) -> str | None:
        return self.settings.password_hash

    async def set_password_hash(self, password_hash: str) -> _FakeSettings:
        self.settings.password_hash = password_hash
        return self.settings

    async def try_set_password_hash(self, password_hash: str) -> bool:
        if self.settings.password_hash is not None:
            return False
        self.settings.password_hash = password_hash
        return True

    async def clear_password_and_totp(self) -> _FakeSettings:
        self.settings.password_hash = None
        self.settings.totp_required_on_login = False
        self.settings.totp_secret_encrypted = None
        self.settings.totp_last_verified_step = None
        return self.settings

    async def set_totp_secret(self, secret_encrypted: bytes | None) -> _FakeSettings:
        self.settings.totp_secret_encrypted = secret_encrypted
        self.settings.totp_last_verified_step = None
        if secret_encrypted is None:
            self.settings.totp_required_on_login = False
        return self.settings

    async def try_advance_totp_last_verified_step(self, step: int) -> bool:
        current = self.settings.totp_last_verified_step
        if current is not None and current >= step:
            return False
        self.settings.totp_last_verified_step = step
        return True


@pytest.mark.asyncio
async def test_setup_password_hashes_and_rejects_duplicate() -> None:
    repository = _FakeRepository()
    service = DashboardAuthService(repository, DashboardSessionStore())

    await service.setup_password("password123")
    stored_hash = repository.settings.password_hash
    assert stored_hash is not None
    assert stored_hash != "password123"
    assert bcrypt.checkpw("password123".encode("utf-8"), stored_hash.encode("utf-8")) is True

    with pytest.raises(PasswordAlreadyConfiguredError):
        await service.setup_password("another-password")


@pytest.mark.asyncio
async def test_setup_password_raises_when_atomic_set_fails() -> None:
    repository = _FakeRepository()
    repository.settings.password_hash = "already-set-by-race"
    service = DashboardAuthService(repository, DashboardSessionStore())

    with pytest.raises(PasswordAlreadyConfiguredError):
        await service.setup_password("password123")


@pytest.mark.asyncio
async def test_verify_and_change_password() -> None:
    repository = _FakeRepository()
    service = DashboardAuthService(repository, DashboardSessionStore())
    await service.setup_password("password123")

    await service.verify_password("password123")
    with pytest.raises(InvalidCredentialsError):
        await service.verify_password("wrong-password")

    await service.change_password("password123", "new-password-456")
    await service.verify_password("new-password-456")
    with pytest.raises(InvalidCredentialsError):
        await service.verify_password("password123")


@pytest.mark.asyncio
async def test_remove_password_clears_password_and_totp() -> None:
    repository = _FakeRepository()
    service = DashboardAuthService(repository, DashboardSessionStore())
    await service.setup_password("password123")
    repository.settings.totp_required_on_login = True
    repository.settings.totp_secret_encrypted = b"secret"
    repository.settings.totp_last_verified_step = 123

    await service.remove_password("password123")
    assert repository.settings.password_hash is None
    assert repository.settings.totp_required_on_login is False
    assert repository.settings.totp_secret_encrypted is None
    assert repository.settings.totp_last_verified_step is None

    with pytest.raises(PasswordNotConfiguredError):
        await service.verify_password("password123")



from __future__ import annotations

from fastapi import APIRouter, Depends, File, Request, UploadFile

from app.core.audit.service import AuditService
from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.exceptions import DashboardBadRequestError, DashboardConflictError, DashboardNotFoundError
from app.dependencies import AccountsContext, get_accounts_context
from app.modules.accounts.repository import AccountIdentityConflictError
from app.modules.accounts.schemas import (
    AccountDeleteResponse,
    AccountImportResponse,
    AccountPauseResponse,
    AccountReactivateResponse,
    AccountsResponse,
    AccountSummary,
    AccountTrendsResponse,
    PlatformIdentityCreateRequest,
    PlatformIdentityUpdateRequest,
)
from app.modules.accounts.service import (
    InvalidAuthJsonError,
    PlatformIdentityConflictError,
    PlatformIdentityNotFoundError,
    PlatformIdentityPrerequisiteError,
)

router = APIRouter(
    prefix="/api/accounts",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("", response_model=AccountsResponse)
async def list_accounts(
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountsResponse:
    accounts = await context.service.list_accounts()
    return AccountsResponse(accounts=accounts)


@router.get("/{account_id}/trends", response_model=AccountTrendsResponse)
async def get_account_trends(
    account_id: str,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountTrendsResponse:
    result = await context.service.get_account_trends(account_id)
    if not result:
        raise DashboardNotFoundError("Account not found", code="account_not_found")
    return result


@router.post("/import", response_model=AccountImportResponse)
async def import_account(
    request: Request,
    auth_json: UploadFile = File(...),
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountImportResponse:
    raw = await auth_json.read()
    try:
        response = await context.service.import_account(raw)
        AuditService.log_async(
            "account_created",
            actor_ip=request.client.host if request.client else None,
            details={"account_id": response.account_id},
        )
        return response
    except InvalidAuthJsonError as exc:
        raise DashboardBadRequestError("Invalid auth.json payload", code="invalid_auth_json") from exc
    except AccountIdentityConflictError as exc:
        raise DashboardConflictError(str(exc), code="duplicate_identity_conflict") from exc


@router.post("/platform", response_model=AccountImportResponse)
async def create_platform_identity(
    request: Request,
    payload: PlatformIdentityCreateRequest,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountImportResponse:
    try:
        response = await context.service.create_platform_identity(payload)
    except PlatformIdentityConflictError as exc:
        raise DashboardConflictError(str(exc), code="platform_identity_conflict") from exc
    except PlatformIdentityPrerequisiteError as exc:
        raise DashboardBadRequestError(str(exc), code="platform_identity_prerequisite_failed") from exc
    except ValueError as exc:
        raise DashboardBadRequestError(str(exc), code="platform_identity_validation_failed") from exc
    AuditService.log_async(
        "platform_identity_created",
        actor_ip=request.client.host if request.client else None,
        details={"account_id": response.account_id},
    )
    return response


@router.patch("/platform/{account_id}", response_model=AccountSummary)
async def update_platform_identity(
    request: Request,
    account_id: str,
    payload: PlatformIdentityUpdateRequest,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountSummary:
    try:
        response = await context.service.update_platform_identity(account_id, payload)
    except PlatformIdentityConflictError as exc:
        raise DashboardConflictError(str(exc), code="platform_identity_conflict") from exc
    except PlatformIdentityNotFoundError as exc:
        raise DashboardNotFoundError(str(exc), code="account_not_found") from exc
    except ValueError as exc:
        raise DashboardBadRequestError(str(exc), code="platform_identity_validation_failed") from exc
    AuditService.log_async(
        "platform_identity_updated",
        actor_ip=request.client.host if request.client else None,
        details={"account_id": account_id},
    )
    return response


@router.post("/{account_id}/reactivate", response_model=AccountReactivateResponse)
async def reactivate_account(
    account_id: str,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountReactivateResponse:
    success = await context.service.reactivate_account(account_id)
    if not success:
        raise DashboardNotFoundError("Account not found", code="account_not_found")
    return AccountReactivateResponse(status="reactivated")


@router.post("/{account_id}/pause", response_model=AccountPauseResponse)
async def pause_account(
    account_id: str,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountPauseResponse:
    success = await context.service.pause_account(account_id)
    if not success:
        raise DashboardNotFoundError("Account not found", code="account_not_found")
    return AccountPauseResponse(status="paused")


@router.delete("/{account_id}", response_model=AccountDeleteResponse)
async def delete_account(
    request: Request,
    account_id: str,
    context: AccountsContext = Depends(get_accounts_context),
) -> AccountDeleteResponse:
    success = await context.service.delete_account(account_id)
    if not success:
        raise DashboardNotFoundError("Account not found", code="account_not_found")
    AuditService.log_async(
        "account_deleted",
        actor_ip=request.client.host if request.client else None,
        details={"account_id": account_id},
    )
    return AccountDeleteResponse(status="deleted")



from __future__ import annotations

import pytest


@pytest.mark.asyncio
@pytest.mark.e2e
async def test_dashboard_password_auth_flow(client, setup_dashboard_password, login_dashboard):
    public_response = await client.get("/api/settings")
    assert public_response.status_code == 200

    password = await setup_dashboard_password(client)

    logout_response = await client.post("/api/dashboard-auth/logout", json={})
    assert logout_response.status_code == 200

    blocked_response = await client.get("/api/settings")
    assert blocked_response.status_code == 401
    assert blocked_response.json()["error"]["code"] == "authentication_required"

    login_payload = await login_dashboard(client, password=password)
    assert login_payload["passwordRequired"] is True
    assert login_payload["authenticated"] is True

    protected_response = await client.get("/api/settings")
    assert protected_response.status_code == 200
    assert protected_response.json()["totpRequiredOnLogin"] is False

    session_response = await client.get("/api/dashboard-auth/session")
    assert session_response.status_code == 200
    assert session_response.json()["authenticated"] is True



from __future__ import annotations

from typing import TypeAlias

from pydantic import (
    BaseModel,
    ConfigDict,
    StrictFloat,
    StrictInt,
    StrictStr,
    ValidationError,
    field_validator,
)


class OpenAIError(BaseModel):
    model_config = ConfigDict(extra="allow")

    message: StrictStr | None = None
    type: StrictStr | None = None
    code: StrictStr | None = None
    param: StrictStr | None = None
    plan_type: StrictStr | None = None
    resets_at: StrictInt | StrictFloat | None = None
    resets_in_seconds: StrictInt | StrictFloat | None = None


class OpenAIErrorEnvelope(BaseModel):
    model_config = ConfigDict(extra="ignore")

    error: OpenAIError | None = None


class ResponseUsageDetails(BaseModel):
    model_config = ConfigDict(extra="allow")

    cached_tokens: StrictInt | None = None
    reasoning_tokens: StrictInt | None = None


class ResponseUsage(BaseModel):
    model_config = ConfigDict(extra="allow")

    input_tokens: StrictInt | None = None
    output_tokens: StrictInt | None = None
    total_tokens: StrictInt | None = None
    input_tokens_details: ResponseUsageDetails | None = None
    output_tokens_details: ResponseUsageDetails | None = None


class OpenAIResponse(BaseModel):
    model_config = ConfigDict(extra="ignore")

    id: StrictStr | None = None
    status: StrictStr | None = None
    error: OpenAIError | None = None
    usage: ResponseUsage | None = None

    @field_validator("error", mode="before")
    @classmethod
    def _normalize_error(cls, value: object) -> OpenAIError | None:
        if value is None:
            return None
        try:
            return OpenAIError.model_validate(value)
        except ValidationError:
            return None

    @field_validator("usage", mode="before")
    @classmethod
    def _normalize_usage(cls, value: object) -> ResponseUsage | None:
        if value is None:
            return None
        try:
            return ResponseUsage.model_validate(value)
        except ValidationError:
            return None


class OpenAIEvent(BaseModel):
    model_config = ConfigDict(extra="ignore")

    type: StrictStr
    response: OpenAIResponse | None = None
    error: OpenAIError | None = None

    @field_validator("error", mode="before")
    @classmethod
    def _normalize_error(cls, value: object) -> OpenAIError | None:
        if value is None:
            return None
        try:
            return OpenAIError.model_validate(value)
        except ValidationError:
            return None


class OpenAIResponsePayload(BaseModel):
    model_config = ConfigDict(extra="allow")

    id: StrictStr | None = None
    status: StrictStr | None = None
    error: OpenAIError | None = None
    usage: ResponseUsage | None = None

    @field_validator("error", mode="before")
    @classmethod
    def _normalize_error(cls, value: object) -> OpenAIError | None:
        if value is None:
            return None
        try:
            return OpenAIError.model_validate(value)
        except ValidationError:
            return None

    @field_validator("usage", mode="before")
    @classmethod
    def _normalize_usage(cls, value: object) -> ResponseUsage | None:
        if value is None:
            return None
        try:
            return ResponseUsage.model_validate(value)
        except ValidationError:
            return None


class CompactResponsePayload(BaseModel):
    model_config = ConfigDict(extra="allow")

    object: StrictStr
    id: StrictStr | None = None
    status: StrictStr | None = None
    error: OpenAIError | None = None
    usage: ResponseUsage | None = None

    @field_validator("object")
    @classmethod
    def _validate_object(cls, value: str) -> str:
        normalized = value.strip()
        if not normalized:
            raise ValueError("Compact response payload requires an object discriminator")
        if not normalized.startswith("response.compact"):
            raise ValueError("Compact response payload requires a compact object discriminator")
        return normalized

    @field_validator("error", mode="before")
    @classmethod
    def _normalize_error(cls, value: object) -> OpenAIError | None:
        if value is None:
            return None
        try:
            return OpenAIError.model_validate(value)
        except ValidationError:
            return None

    @field_validator("usage", mode="before")
    @classmethod
    def _normalize_usage(cls, value: object) -> ResponseUsage | None:
        if value is None:
            return None
        try:
            return ResponseUsage.model_validate(value)
        except ValidationError:
            return None


OpenAIResponseResult: TypeAlias = OpenAIResponsePayload | OpenAIErrorEnvelope
CompactResponseResult: TypeAlias = CompactResponsePayload | OpenAIErrorEnvelope



from __future__ import annotations

from urllib.parse import quote

import pytest

pytestmark = pytest.mark.integration


@pytest.mark.asyncio
async def test_firewall_api_crud_and_mode(async_client):
    initial = await async_client.get("/api/firewall/ips")
    assert initial.status_code == 200
    initial_payload = initial.json()
    assert initial_payload["mode"] == "allow_all"
    assert initial_payload["entries"] == []

    created = await async_client.post("/api/firewall/ips", json={"ipAddress": "127.0.0.1"})
    assert created.status_code == 200
    created_payload = created.json()
    assert created_payload["ipAddress"] == "127.0.0.1"
    assert isinstance(created_payload["createdAt"], str)

    listed = await async_client.get("/api/firewall/ips")
    assert listed.status_code == 200
    listed_payload = listed.json()
    assert listed_payload["mode"] == "allowlist_active"
    assert len(listed_payload["entries"]) == 1
    assert listed_payload["entries"][0]["ipAddress"] == "127.0.0.1"

    duplicate = await async_client.post("/api/firewall/ips", json={"ipAddress": "127.0.0.1"})
    assert duplicate.status_code == 409
    duplicate_payload = duplicate.json()
    assert duplicate_payload["error"]["code"] == "ip_exists"

    invalid = await async_client.post("/api/firewall/ips", json={"ipAddress": "not-an-ip"})
    assert invalid.status_code == 400
    invalid_payload = invalid.json()
    assert invalid_payload["error"]["code"] == "invalid_ip"

    deleted = await async_client.delete("/api/firewall/ips/127.0.0.1")
    assert deleted.status_code == 200
    assert deleted.json()["status"] == "deleted"

    missing = await async_client.delete("/api/firewall/ips/127.0.0.1")
    assert missing.status_code == 404
    missing_payload = missing.json()
    assert missing_payload["error"]["code"] == "ip_not_found"

    final = await async_client.get("/api/firewall/ips")
    assert final.status_code == 200
    final_payload = final.json()
    assert final_payload["mode"] == "allow_all"
    assert final_payload["entries"] == []


@pytest.mark.asyncio
async def test_firewall_api_ipv6_normalized(async_client):
    created = await async_client.post(
        "/api/firewall/ips",
        json={"ipAddress": "2001:0db8:0000:0000:0000:ff00:0042:8329"},
    )
    assert created.status_code == 200
    payload = created.json()
    assert payload["ipAddress"] == "2001:db8::ff00:42:8329"

    encoded = quote(payload["ipAddress"], safe="")
    deleted = await async_client.delete(f"/api/firewall/ips/{encoded}")
    assert deleted.status_code == 200



"""add dashboard upstream stream transport setting

Revision ID: 20260312_120000_add_dashboard_upstream_stream_transport
Revises: 20260312_000000_split_sticky_sessions_primary_key_by_kind
Create Date: 2026-03-12
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260312_120000_add_dashboard_upstream_stream_transport"
down_revision = "20260313_024500_merge_request_log_transport_heads"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(column["name"]) for column in inspector.get_columns(table_name) if column.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "dashboard_settings"):
        return
    if "upstream_stream_transport" in _columns(bind, "dashboard_settings"):
        return

    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.add_column(
            sa.Column(
                "upstream_stream_transport",
                sa.String(),
                nullable=False,
                server_default=sa.text("'default'"),
            )
        )

    bind.execute(
        sa.text(
            """
            UPDATE dashboard_settings
            SET upstream_stream_transport = 'default'
            WHERE upstream_stream_transport IS NULL OR upstream_stream_transport = ''
            """
        )
    )


def downgrade() -> None:
    return



from __future__ import annotations

import asyncio
import contextlib
import importlib
import logging
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Protocol, cast

from app.core.config.settings import get_settings
from app.core.utils.time import utcnow
from app.db.session import get_background_session
from app.modules.proxy.sticky_repository import StickySessionsRepository
from app.modules.settings.repository import SettingsRepository

logger = logging.getLogger(__name__)


class _LeaderElectionLike(Protocol):
    async def try_acquire(self) -> bool: ...


def _get_leader_election() -> _LeaderElectionLike:
    module = importlib.import_module("app.core.scheduling.leader_election")
    return cast(_LeaderElectionLike, module.get_leader_election())


@dataclass(slots=True)
class StickySessionCleanupScheduler:
    interval_seconds: int
    enabled: bool
    _task: asyncio.Task[None] | None = None
    _stop: asyncio.Event = field(default_factory=asyncio.Event)
    _lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    async def start(self) -> None:
        if not self.enabled:
            return
        if self._task and not self._task.done():
            return
        self._stop.clear()
        self._task = asyncio.create_task(self._run_loop())

    async def stop(self) -> None:
        if not self._task:
            return
        self._stop.set()
        self._task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._task
        self._task = None

    async def _run_loop(self) -> None:
        while not self._stop.is_set():
            await self._cleanup_once()
            try:
                await asyncio.wait_for(self._stop.wait(), timeout=self.interval_seconds)
            except asyncio.TimeoutError:
                continue

    async def _cleanup_once(self) -> None:
        if not await _get_leader_election().try_acquire():
            return
        async with self._lock:
            try:
                async with get_background_session() as session:
                    settings_repo = SettingsRepository(session)
                    sticky_repo = StickySessionsRepository(session)
                    settings = await settings_repo.get_or_create()

                    cutoff = utcnow() - timedelta(seconds=settings.openai_cache_affinity_max_age_seconds)
                    deleted_count = await sticky_repo.purge_prompt_cache_before(cutoff)
                    if deleted_count > 0:
                        logger.info("Purged stale prompt-cache sticky sessions deleted_count=%s", deleted_count)
            except Exception:
                logger.exception("Sticky session cleanup loop failed")


def build_sticky_session_cleanup_scheduler() -> StickySessionCleanupScheduler:
    settings = get_settings()
    return StickySessionCleanupScheduler(
        interval_seconds=settings.sticky_session_cleanup_interval_seconds,
        enabled=settings.sticky_session_cleanup_enabled,
    )



"""merge scheduler leader and cache locality heads

Revision ID: 20260330_010000_merge_scheduler_leader_and_cache_locality_heads
Revises: 20260328_140000_add_scheduler_leader_table, 20260330_000000_add_cache_locality_settings
Create Date: 2026-03-30
"""

from __future__ import annotations

# revision identifiers, used by Alembic.
revision = "20260330_010000_merge_scheduler_leader_and_cache_locality_heads"
down_revision = (
    "20260328_140000_add_scheduler_leader_table",
    "20260330_000000_add_cache_locality_settings",
)
branch_labels = None
depends_on = None


def upgrade() -> None:
    return


def downgrade() -> None:
    return



from __future__ import annotations

from collections.abc import Mapping
from typing import TypeGuard

from app.core.types import JsonValue


def is_json_mapping(value: object) -> TypeGuard[Mapping[str, JsonValue]]:
    return isinstance(value, Mapping)


def is_json_dict(value: object) -> TypeGuard[dict[str, JsonValue]]:
    return isinstance(value, dict)


def is_json_list(value: object) -> TypeGuard[list[JsonValue]]:
    return isinstance(value, list)



from __future__ import annotations

from fastapi import APIRouter, Body, Depends

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.cache.invalidation import NAMESPACE_FIREWALL, get_cache_invalidation_poller
from app.core.exceptions import DashboardBadRequestError, DashboardConflictError, DashboardNotFoundError
from app.core.middleware.firewall_cache import get_firewall_ip_cache
from app.dependencies import FirewallContext, get_firewall_context
from app.modules.firewall.schemas import (
    FirewallDeleteResponse,
    FirewallIpCreateRequest,
    FirewallIpEntry,
    FirewallIpsResponse,
)
from app.modules.firewall.service import FirewallIpAlreadyExistsError, FirewallValidationError

router = APIRouter(
    prefix="/api/firewall",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("/ips", response_model=FirewallIpsResponse)
async def list_firewall_ips(
    context: FirewallContext = Depends(get_firewall_context),
) -> FirewallIpsResponse:
    payload = await context.service.list_ips()
    return FirewallIpsResponse(
        mode=payload.mode,
        entries=[
            FirewallIpEntry(ip_address=entry.ip_address, created_at=entry.created_at) for entry in payload.entries
        ],
    )


@router.post("/ips", response_model=FirewallIpEntry)
async def add_firewall_ip(
    payload: FirewallIpCreateRequest = Body(...),
    context: FirewallContext = Depends(get_firewall_context),
) -> FirewallIpEntry:
    try:
        created = await context.service.add_ip(payload.ip_address)
    except FirewallValidationError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_ip") from exc
    except FirewallIpAlreadyExistsError as exc:
        raise DashboardConflictError(str(exc), code="ip_exists") from exc
    get_firewall_ip_cache().invalidate_all()
    poller = get_cache_invalidation_poller()
    if poller is not None:
        await poller.bump(NAMESPACE_FIREWALL)
    return FirewallIpEntry(ip_address=created.ip_address, created_at=created.created_at)


@router.delete("/ips/{ip_address}", response_model=FirewallDeleteResponse)
async def delete_firewall_ip(
    ip_address: str,
    context: FirewallContext = Depends(get_firewall_context),
) -> FirewallDeleteResponse:
    try:
        deleted = await context.service.remove_ip(ip_address)
    except FirewallValidationError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_ip") from exc
    if not deleted:
        raise DashboardNotFoundError("IP address not found", code="ip_not_found")
    get_firewall_ip_cache().invalidate_all()
    poller = get_cache_invalidation_poller()
    if poller is not None:
        await poller.bump(NAMESPACE_FIREWALL)
    return FirewallDeleteResponse(status="deleted")



"""switch import_without_overwrite default to true

Revision ID: 20260408_000000_switch_import_without_overwrite_default_to_true
Revises: 20260403_000000_add_credit_api_key_limit_values
Create Date: 2026-04-08
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

revision = "20260408_000000_switch_import_without_overwrite_default_to_true"
down_revision = "20260403_000000_add_credit_api_key_limit_values"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(column["name"]) for column in inspector.get_columns(table_name) if column.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "dashboard_settings"):
        return
    columns = _columns(bind, "dashboard_settings")
    if "import_without_overwrite" not in columns:
        return

    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.alter_column(
            "import_without_overwrite",
            existing_type=sa.Boolean(),
            server_default=sa.true(),
        )

    if {"created_at", "updated_at"}.issubset(columns):
        op.execute(
            sa.text(
                """
                UPDATE dashboard_settings
                SET import_without_overwrite = TRUE
                WHERE import_without_overwrite = FALSE
                  AND updated_at = created_at
                """
            )
        )


def downgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "dashboard_settings"):
        return
    columns = _columns(bind, "dashboard_settings")
    if "import_without_overwrite" not in columns:
        return

    with op.batch_alter_table("dashboard_settings") as batch_op:
        batch_op.alter_column(
            "import_without_overwrite",
            existing_type=sa.Boolean(),
            server_default=sa.false(),
        )



#!/usr/bin/env python3
"""UserPromptSubmit hook — suggest relevant skills based on prompt keywords/intent."""

import json
import re
import sys
from pathlib import Path

try:
    from _analytics import emit_event
except ImportError:

    def emit_event(*_a: object, **_k: object) -> None:
        pass


def load_skill_rules() -> dict:
    """Load skill-rules.json relative to this hook's location."""
    hook_dir = Path(__file__).resolve().parent
    rules_path = hook_dir.parent / "skills" / "skill-rules.json"
    with open(rules_path) as f:
        return json.load(f)


def match_prompt_triggers(prompt: str, triggers: dict) -> bool:
    """Check if prompt matches keyword or intent pattern triggers."""
    prompt_lower = prompt.lower()

    for kw in triggers.get("keywords", []):
        if kw.lower() in prompt_lower:
            return True

    for pattern in triggers.get("intentPatterns", []):
        if re.search(pattern, prompt, re.IGNORECASE):
            return True

    return False


def main() -> None:
    try:
        input_data = json.loads(sys.stdin.read())
    except (json.JSONDecodeError, EOFError):
        sys.exit(0)

    prompt = input_data.get("prompt", "")
    if not prompt.strip():
        sys.exit(0)

    try:
        rules = load_skill_rules()
    except (FileNotFoundError, json.JSONDecodeError):
        sys.exit(0)

    skills = rules.get("skills", {})

    # Match each skill's promptTriggers against the user prompt
    matched: list[dict] = []
    for name, rule in skills.items():
        triggers = rule.get("promptTriggers")
        if not triggers:
            continue
        if match_prompt_triggers(prompt, triggers):
            matched.append(
                {
                    "name": name,
                    "priority": rule.get("priority", "medium"),
                    "description": rule.get("description", ""),
                }
            )

    if not matched:
        emit_event(
            input_data.get("session_id", "unknown"),
            "hook.invoked",
            {
                "hook": "skill_activation",
                "trigger": "UserPromptSubmit",
                "outcome": "no_match",
                "matched_count": 0,
                "exit_code": 0,
            },
        )
        sys.exit(0)

    # Group by priority
    priority_order = ["critical", "high", "medium", "low"]
    by_priority: dict[str, list[dict]] = {p: [] for p in priority_order}
    for m in matched:
        by_priority.setdefault(m["priority"], []).append(m)

    # Build output banner
    lines: list[str] = [
        "",
        "━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━",
        "SKILL ACTIVATION CHECK",
        "━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━",
        "",
    ]

    labels = {
        "critical": "CRITICAL SKILLS (REQUIRED):",
        "high": "RECOMMENDED SKILLS:",
        "medium": "SUGGESTED SKILLS:",
        "low": "OPTIONAL SKILLS:",
    }

    for priority in priority_order:
        group = by_priority.get(priority, [])
        if not group:
            continue
        lines.append(f"  {labels[priority]}")
        for s in group:
            lines.append(f"    -> {s['name']}")
        lines.append("")

    lines.append("  ACTION: Use Skill tool to invoke matched skills")
    lines.append("━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━")
    lines.append("")

    emit_event(
        input_data.get("session_id", "unknown"),
        "hook.invoked",
        {
            "hook": "skill_activation",
            "trigger": "UserPromptSubmit",
            "outcome": "matched",
            "matched_count": len(matched),
            "exit_code": 0,
        },
    )
    print("\n".join(lines))
    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import uuid
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from hashlib import sha256
from typing import TYPE_CHECKING

from sqlalchemy import delete, select, update
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.utils.time import utcnow
from app.db.models import BridgeRingMember

if TYPE_CHECKING:
    from collections.abc import Callable


RING_HEARTBEAT_INTERVAL_SECONDS = 30
RING_STALE_THRESHOLD_SECONDS = 120
RING_STALE_GRACE_SECONDS = RING_HEARTBEAT_INTERVAL_SECONDS + 5
_SQLITE_URL_PREFIXES = ("sqlite+aiosqlite:///", "sqlite:///")


def dynamic_bridge_ring_membership_enabled(settings: object) -> bool:
    bridge_enabled = getattr(settings, "http_responses_session_bridge_enabled", True)
    if isinstance(bridge_enabled, bool) and not bridge_enabled:
        return False

    database_url = getattr(settings, "database_url", None)
    if isinstance(database_url, str) and database_url.startswith(_SQLITE_URL_PREFIXES):
        return False

    ring_value = getattr(settings, "http_responses_session_bridge_instance_ring", None)
    if isinstance(ring_value, list):
        normalized = []
        for entry in ring_value:
            if isinstance(entry, str):
                stripped = entry.strip()
                if stripped:
                    normalized.append(stripped)
        if normalized:
            return False

    return True


class RingMembershipService:
    """Manages pod registration in the bridge ring.

    This service stores and retrieves active pod memberships from the DB,
    ensuring all pods see the same ring view (solving the split-brain problem).
    """

    def __init__(self, session_factory: Callable[[], AsyncSession]) -> None:
        self._session_factory = session_factory

    async def register(self, instance_id: str) -> None:
        """Upsert pod into ring. Safe to call multiple times."""
        async with self._session() as session:
            # Dialect-specific upsert
            dialect = session.get_bind().dialect.name
            if dialect == "postgresql":
                stmt = (
                    pg_insert(BridgeRingMember)
                    .values(
                        id=str(uuid.uuid4()),
                        instance_id=instance_id,
                        registered_at=utcnow(),
                        last_heartbeat_at=utcnow(),
                    )
                    .on_conflict_do_update(
                        index_elements=["instance_id"],
                        set_={"last_heartbeat_at": utcnow(), "registered_at": utcnow()},
                    )
                )
            elif dialect == "sqlite":
                stmt = (
                    sqlite_insert(BridgeRingMember)
                    .values(
                        id=str(uuid.uuid4()),
                        instance_id=instance_id,
                        registered_at=utcnow(),
                        last_heartbeat_at=utcnow(),
                    )
                    .on_conflict_do_update(
                        index_elements=["instance_id"],
                        set_={"last_heartbeat_at": utcnow(), "registered_at": utcnow()},
                    )
                )
            else:
                raise RuntimeError(f"RingMembershipService unsupported for dialect={dialect!r}")
            await session.execute(stmt)
            await session.commit()

    async def heartbeat(self, instance_id: str) -> None:
        """Upsert heartbeat — recovers from mark_stale or unregister by sibling workers."""
        async with self._session() as session:
            dialect = session.get_bind().dialect.name
            now = utcnow()
            if dialect == "postgresql":
                stmt = (
                    pg_insert(BridgeRingMember)
                    .values(
                        id=str(uuid.uuid4()),
                        instance_id=instance_id,
                        registered_at=now,
                        last_heartbeat_at=now,
                    )
                    .on_conflict_do_update(
                        index_elements=["instance_id"],
                        set_={"last_heartbeat_at": now},
                    )
                )
            elif dialect == "sqlite":
                stmt = (
                    sqlite_insert(BridgeRingMember)
                    .values(
                        id=str(uuid.uuid4()),
                        instance_id=instance_id,
                        registered_at=now,
                        last_heartbeat_at=now,
                    )
                    .on_conflict_do_update(
                        index_elements=["instance_id"],
                        set_={"last_heartbeat_at": now},
                    )
                )
            else:
                stmt = (
                    update(BridgeRingMember)
                    .where(BridgeRingMember.instance_id == instance_id)
                    .values(last_heartbeat_at=now)
                )
            await session.execute(stmt)
            await session.commit()

    async def unregister(self, instance_id: str) -> None:
        """Remove pod from ring."""
        async with self._session() as session:
            stmt = delete(BridgeRingMember).where(BridgeRingMember.instance_id == instance_id)
            await session.execute(stmt)
            await session.commit()

    async def mark_stale(
        self,
        instance_id: str,
        *,
        stale_threshold_seconds: int = RING_STALE_THRESHOLD_SECONDS,
        grace_seconds: int = RING_STALE_GRACE_SECONDS,
    ) -> None:
        """Age the heartbeat close to expiry without deleting the shared row.

        A short grace window lets sibling workers refresh the shared row on
        their next heartbeat, while a fully terminating pod still ages out far
        faster than the normal stale threshold.
        """
        from datetime import timedelta

        active_for_seconds = max(grace_seconds, 0)
        age_seconds = max(stale_threshold_seconds - active_for_seconds, 0)
        stale_time = utcnow() - timedelta(seconds=age_seconds)
        async with self._session() as session:
            stmt = (
                update(BridgeRingMember)
                .where(BridgeRingMember.instance_id == instance_id)
                .values(last_heartbeat_at=stale_time)
            )
            await session.execute(stmt)
            await session.commit()

    async def list_active(self, stale_threshold_seconds: int = RING_STALE_THRESHOLD_SECONDS) -> list[str]:
        """Return sorted list of pods whose heartbeat is within threshold."""
        from datetime import timedelta

        cutoff = utcnow() - timedelta(seconds=stale_threshold_seconds)
        async with self._session() as session:
            result = await session.execute(
                select(BridgeRingMember.instance_id)
                .where(BridgeRingMember.last_heartbeat_at >= cutoff)
                .order_by(BridgeRingMember.instance_id)
            )
            return list(result.scalars().all())

    async def ring_fingerprint(self, stale_threshold_seconds: int = RING_STALE_THRESHOLD_SECONDS) -> str:
        """sha256 of sorted active member list. Same for all pods with same membership."""
        members = await self.list_active(stale_threshold_seconds)
        data = ",".join(sorted(members))
        return sha256(data.encode()).hexdigest()

    @asynccontextmanager
    async def _session(self) -> AsyncIterator[AsyncSession]:
        session = self._session_factory()
        try:
            yield session
        finally:
            await session.close()



from __future__ import annotations

from pydantic import Field

from app.modules.shared.schemas import DashboardModel


class DashboardSettingsResponse(DashboardModel):
    sticky_threads_enabled: bool
    upstream_stream_transport: str = Field(pattern=r"^(default|auto|http|websocket)$")
    prefer_earlier_reset_accounts: bool
    routing_strategy: str = Field(pattern=r"^(usage_weighted|round_robin|capacity_weighted)$")
    openai_cache_affinity_max_age_seconds: int = Field(gt=0)
    http_responses_session_bridge_prompt_cache_idle_ttl_seconds: int = Field(gt=0)
    http_responses_session_bridge_gateway_safe_mode: bool
    sticky_reallocation_budget_threshold_pct: float = Field(ge=0.0, le=100.0)
    import_without_overwrite: bool
    totp_required_on_login: bool
    totp_configured: bool
    api_key_auth_enabled: bool


class DashboardSettingsUpdateRequest(DashboardModel):
    sticky_threads_enabled: bool
    upstream_stream_transport: str | None = Field(
        default=None,
        pattern=r"^(default|auto|http|websocket)$",
    )
    prefer_earlier_reset_accounts: bool
    routing_strategy: str | None = Field(default=None, pattern=r"^(usage_weighted|round_robin|capacity_weighted)$")
    openai_cache_affinity_max_age_seconds: int | None = Field(default=None, gt=0)
    http_responses_session_bridge_prompt_cache_idle_ttl_seconds: int | None = Field(default=None, gt=0)
    http_responses_session_bridge_gateway_safe_mode: bool | None = None
    sticky_reallocation_budget_threshold_pct: float | None = Field(default=None, ge=0.0, le=100.0)
    import_without_overwrite: bool | None = None
    totp_required_on_login: bool | None = None
    api_key_auth_enabled: bool | None = None


class RuntimeConnectAddressResponse(DashboardModel):
    connect_address: str



from __future__ import annotations

import ipaddress
import os
import socket

from fastapi import APIRouter, Body, Depends, Request

from app.core.audit.service import AuditService
from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.config.settings_cache import get_settings_cache
from app.core.exceptions import DashboardBadRequestError
from app.dependencies import SettingsContext, get_settings_context
from app.modules.settings.schemas import (
    DashboardSettingsResponse,
    DashboardSettingsUpdateRequest,
    RuntimeConnectAddressResponse,
)
from app.modules.settings.service import DashboardSettingsUpdateData

LOOPBACK_HOSTS = {"localhost", "127.0.0.1", "::1", "[::1]"}


def _is_non_loopback_ipv4(value: str | None) -> bool:
    if not value:
        return False
    try:
        address = ipaddress.ip_address(value)
    except ValueError:
        return False
    return isinstance(address, ipaddress.IPv4Address) and not address.is_loopback and not address.is_unspecified


def _resolve_hostname_ipv4(hostname: str) -> str | None:
    try:
        infos = socket.getaddrinfo(hostname, None, family=socket.AF_INET, type=socket.SOCK_STREAM)
    except OSError:
        return None
    for info in infos:
        candidate = info[4][0]
        if not isinstance(candidate, str):
            continue
        if _is_non_loopback_ipv4(candidate):
            return candidate
    return None


def _resolve_runtime_connect_address(request: Request) -> str:
    override = os.getenv("CODEX_LB_CONNECT_ADDRESS", "").strip()
    if override:
        return override

    request_host = request.url.hostname or ""
    if _is_non_loopback_ipv4(request_host):
        return request_host

    normalized_host = request_host.strip().lower()
    if normalized_host and normalized_host not in LOOPBACK_HOSTS:
        resolved_host = _resolve_hostname_ipv4(request_host)
        if resolved_host:
            return resolved_host
        return request_host
    return "<codex-lb-ip-or-dns>"


router = APIRouter(
    prefix="/api/settings",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("", response_model=DashboardSettingsResponse)
async def get_settings(
    context: SettingsContext = Depends(get_settings_context),
) -> DashboardSettingsResponse:
    settings = await context.service.get_settings()
    return DashboardSettingsResponse(
        sticky_threads_enabled=settings.sticky_threads_enabled,
        upstream_stream_transport=settings.upstream_stream_transport,
        prefer_earlier_reset_accounts=settings.prefer_earlier_reset_accounts,
        routing_strategy=settings.routing_strategy,
        openai_cache_affinity_max_age_seconds=settings.openai_cache_affinity_max_age_seconds,
        http_responses_session_bridge_prompt_cache_idle_ttl_seconds=settings.http_responses_session_bridge_prompt_cache_idle_ttl_seconds,
        http_responses_session_bridge_gateway_safe_mode=settings.http_responses_session_bridge_gateway_safe_mode,
        sticky_reallocation_budget_threshold_pct=settings.sticky_reallocation_budget_threshold_pct,
        import_without_overwrite=settings.import_without_overwrite,
        totp_required_on_login=settings.totp_required_on_login,
        totp_configured=settings.totp_configured,
        api_key_auth_enabled=settings.api_key_auth_enabled,
    )


@router.get("/runtime/connect-address", response_model=RuntimeConnectAddressResponse)
async def get_runtime_connect_address(request: Request) -> RuntimeConnectAddressResponse:
    return RuntimeConnectAddressResponse(connect_address=_resolve_runtime_connect_address(request))


@router.put("", response_model=DashboardSettingsResponse)
async def update_settings(
    request: Request,
    payload: DashboardSettingsUpdateRequest = Body(...),
    context: SettingsContext = Depends(get_settings_context),
) -> DashboardSettingsResponse:
    current = await context.service.get_settings()
    try:
        updated = await context.service.update_settings(
            DashboardSettingsUpdateData(
                sticky_threads_enabled=payload.sticky_threads_enabled,
                upstream_stream_transport=payload.upstream_stream_transport or current.upstream_stream_transport,
                prefer_earlier_reset_accounts=payload.prefer_earlier_reset_accounts,
                routing_strategy=payload.routing_strategy or current.routing_strategy,
                openai_cache_affinity_max_age_seconds=(
                    payload.openai_cache_affinity_max_age_seconds
                    if payload.openai_cache_affinity_max_age_seconds is not None
                    else current.openai_cache_affinity_max_age_seconds
                ),
                http_responses_session_bridge_prompt_cache_idle_ttl_seconds=(
                    payload.http_responses_session_bridge_prompt_cache_idle_ttl_seconds
                    if payload.http_responses_session_bridge_prompt_cache_idle_ttl_seconds is not None
                    else current.http_responses_session_bridge_prompt_cache_idle_ttl_seconds
                ),
                http_responses_session_bridge_gateway_safe_mode=(
                    payload.http_responses_session_bridge_gateway_safe_mode
                    if payload.http_responses_session_bridge_gateway_safe_mode is not None
                    else current.http_responses_session_bridge_gateway_safe_mode
                ),
                sticky_reallocation_budget_threshold_pct=(
                    payload.sticky_reallocation_budget_threshold_pct
                    if payload.sticky_reallocation_budget_threshold_pct is not None
                    else current.sticky_reallocation_budget_threshold_pct
                ),
                import_without_overwrite=(
                    payload.import_without_overwrite
                    if payload.import_without_overwrite is not None
                    else current.import_without_overwrite
                ),
                totp_required_on_login=(
                    payload.totp_required_on_login
                    if payload.totp_required_on_login is not None
                    else current.totp_required_on_login
                ),
                api_key_auth_enabled=(
                    payload.api_key_auth_enabled
                    if payload.api_key_auth_enabled is not None
                    else current.api_key_auth_enabled
                ),
            )
        )
    except ValueError as exc:
        raise DashboardBadRequestError(str(exc), code="invalid_totp_config") from exc

    await get_settings_cache().invalidate()
    changed_fields = [
        field_name
        for field_name in (
            "sticky_threads_enabled",
            "upstream_stream_transport",
            "prefer_earlier_reset_accounts",
            "routing_strategy",
            "openai_cache_affinity_max_age_seconds",
            "http_responses_session_bridge_gateway_safe_mode",
            "import_without_overwrite",
            "totp_required_on_login",
            "api_key_auth_enabled",
        )
        if getattr(current, field_name) != getattr(updated, field_name)
    ]
    AuditService.log_async(
        "settings_changed",
        actor_ip=request.client.host if request.client else None,
        details={"changed_fields": changed_fields},
    )
    return DashboardSettingsResponse(
        sticky_threads_enabled=updated.sticky_threads_enabled,
        upstream_stream_transport=updated.upstream_stream_transport,
        prefer_earlier_reset_accounts=updated.prefer_earlier_reset_accounts,
        routing_strategy=updated.routing_strategy,
        openai_cache_affinity_max_age_seconds=updated.openai_cache_affinity_max_age_seconds,
        http_responses_session_bridge_prompt_cache_idle_ttl_seconds=updated.http_responses_session_bridge_prompt_cache_idle_ttl_seconds,
        http_responses_session_bridge_gateway_safe_mode=updated.http_responses_session_bridge_gateway_safe_mode,
        sticky_reallocation_budget_threshold_pct=updated.sticky_reallocation_budget_threshold_pct,
        import_without_overwrite=updated.import_without_overwrite,
        totp_required_on_login=updated.totp_required_on_login,
        totp_configured=updated.totp_configured,
        api_key_auth_enabled=updated.api_key_auth_enabled,
    )



from __future__ import annotations

from fastapi import APIRouter, Body, Depends, Query

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.core.exceptions import DashboardNotFoundError
from app.db.models import StickySessionKind
from app.dependencies import StickySessionsContext, get_sticky_sessions_context
from app.modules.sticky_sessions.schemas import (
    StickySessionDeleteFailure,
    StickySessionDeleteResponse,
    StickySessionEntryResponse,
    StickySessionIdentifier,
    StickySessionsDeleteFilteredRequest,
    StickySessionsDeleteFilteredResponse,
    StickySessionsDeleteRequest,
    StickySessionsDeleteResponse,
    StickySessionsListResponse,
    StickySessionSortBy,
    StickySessionSortDir,
    StickySessionsPurgeRequest,
    StickySessionsPurgeResponse,
)
from app.modules.upstream_identities.types import CHATGPT_WEB_PROVIDER_KIND, ProviderKind

router = APIRouter(
    prefix="/api/sticky-sessions",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("", response_model=StickySessionsListResponse)
async def list_sticky_sessions(
    kind: StickySessionKind | None = Query(default=None),
    stale_only: bool = Query(default=False, alias="staleOnly"),
    provider_kind: ProviderKind | None = Query(default=None, alias="providerKind"),
    account_query: str | None = Query(default=None, alias="accountQuery"),
    key_query: str | None = Query(default=None, alias="keyQuery"),
    sort_by: StickySessionSortBy = Query(default="updated_at", alias="sortBy"),
    sort_dir: StickySessionSortDir = Query(default="desc", alias="sortDir"),
    offset: int = Query(default=0, ge=0),
    limit: int = Query(default=100, ge=1, le=500),
    context: StickySessionsContext = Depends(get_sticky_sessions_context),
) -> StickySessionsListResponse:
    result = await context.service.list_entries(
        kind=kind,
        stale_only=stale_only,
        provider_kind=provider_kind,
        account_query=account_query,
        key_query=key_query,
        sort_by=sort_by,
        sort_dir=sort_dir,
        offset=offset,
        limit=limit,
    )
    return StickySessionsListResponse(
        entries=[
            StickySessionEntryResponse(
                key=entry.key,
                display_name=entry.display_name,
                kind=entry.kind,
                provider_kind=entry.provider_kind,
                routing_subject_id=entry.routing_subject_id,
                affinity_scope=entry.affinity_scope,
                created_at=entry.created_at,
                updated_at=entry.updated_at,
                expires_at=entry.expires_at,
                is_stale=entry.is_stale,
            )
            for entry in result.entries
        ],
        stale_prompt_cache_count=result.stale_prompt_cache_count,
        total=result.total,
        has_more=result.has_more,
    )


@router.post("/purge", response_model=StickySessionsPurgeResponse)
async def purge_sticky_sessions(
    payload: StickySessionsPurgeRequest = Body(default=StickySessionsPurgeRequest()),
    context: StickySessionsContext = Depends(get_sticky_sessions_context),
) -> StickySessionsPurgeResponse:
    deleted_count = await context.service.purge_entries()
    return StickySessionsPurgeResponse(deleted_count=deleted_count)


@router.post("/delete", response_model=StickySessionsDeleteResponse)
async def delete_sticky_sessions(
    payload: StickySessionsDeleteRequest,
    context: StickySessionsContext = Depends(get_sticky_sessions_context),
) -> StickySessionsDeleteResponse:
    result = await context.service.delete_entries(
        [(entry.key, entry.kind, entry.provider_kind) for entry in payload.sessions]
    )
    return StickySessionsDeleteResponse(
        deleted_count=result.deleted_count,
        deleted=[
            StickySessionIdentifier(key=key, kind=kind, provider_kind=provider_kind)
            for key, kind, provider_kind in result.deleted
        ],
        failed=[
            StickySessionDeleteFailure(
                key=entry.key,
                kind=entry.kind,
                provider_kind=entry.provider_kind,
                reason=entry.reason,
            )
            for entry in result.failed
        ],
    )


@router.post("/delete-filtered", response_model=StickySessionsDeleteFilteredResponse)
async def delete_filtered_sticky_sessions(
    payload: StickySessionsDeleteFilteredRequest,
    context: StickySessionsContext = Depends(get_sticky_sessions_context),
) -> StickySessionsDeleteFilteredResponse:
    deleted_count = await context.service.delete_filtered_entries(
        stale_only=payload.stale_only,
        provider_kind=payload.provider_kind,
        account_query=payload.account_query,
        key_query=payload.key_query,
    )
    return StickySessionsDeleteFilteredResponse(deleted_count=deleted_count)


@router.delete("/{kind}/{key:path}", response_model=StickySessionDeleteResponse)
async def delete_sticky_session(
    kind: StickySessionKind,
    key: str,
    provider_kind: ProviderKind = Query(default=CHATGPT_WEB_PROVIDER_KIND, alias="providerKind"),
    context: StickySessionsContext = Depends(get_sticky_sessions_context),
) -> StickySessionDeleteResponse:
    deleted = await context.service.delete_entry(key, kind=kind, provider_kind=provider_kind)
    if not deleted:
        raise DashboardNotFoundError("Sticky session not found", code="sticky_session_not_found")
    return StickySessionDeleteResponse(status="deleted")



"""merge additional usage and sticky session heads

Revision ID: 20260312_010000_merge_additional_usage_and_sticky_session_heads
Revises: 20260312_000000_add_additional_usage_quota_key, 20260312_000000_split_sticky_sessions_primary_key_by_kind
Create Date: 2026-03-12
"""

from __future__ import annotations

# revision identifiers, used by Alembic.
revision = "20260312_010000_merge_additional_usage_and_sticky_session_heads"
down_revision = (
    "20260312_000000_add_additional_usage_quota_key",
    "20260312_000000_split_sticky_sessions_primary_key_by_kind",
)
branch_labels = None
depends_on = None


def upgrade() -> None:
    return


def downgrade() -> None:
    return



from __future__ import annotations

import time

import anyio

from app.db.models import DashboardSettings
from app.db.session import SessionLocal
from app.modules.settings.repository import SettingsRepository


class SettingsCache:
    def __init__(self, *, ttl_seconds: float = 5.0) -> None:
        if ttl_seconds <= 0:
            raise ValueError("ttl_seconds must be positive")
        self._ttl_seconds = ttl_seconds
        self._cached_settings: DashboardSettings | None = None
        self._cached_at = 0.0
        self._lock = anyio.Lock()

    async def get(self) -> DashboardSettings:
        now = time.monotonic()
        if self._cached_settings is not None and now - self._cached_at < self._ttl_seconds:
            return self._cached_settings

        async with self._lock:
            now = time.monotonic()
            if self._cached_settings is not None and now - self._cached_at < self._ttl_seconds:
                return self._cached_settings

            async with SessionLocal() as session:
                settings = await SettingsRepository(session).get_or_create()
                self._cached_settings = settings
                self._cached_at = now
                return settings

    async def invalidate(self) -> None:
        async with self._lock:
            self._cached_settings = None
            self._cached_at = 0.0


_settings_cache = SettingsCache()


def get_settings_cache() -> SettingsCache:
    return _settings_cache



from __future__ import annotations

import sqlite3
from dataclasses import dataclass
from enum import Enum
from pathlib import Path


@dataclass(slots=True)
class IntegrityCheck:
    ok: bool
    details: str | None


class SqliteIntegrityCheckMode(str, Enum):
    QUICK = "quick"
    FULL = "full"


def sqlite_db_path_from_url(url: str) -> Path | None:
    if not (url.startswith("sqlite+aiosqlite:") or url.startswith("sqlite:")):
        return None

    marker = ":///"
    marker_index = url.find(marker)
    if marker_index < 0:
        return None

    path = url[marker_index + len(marker) :]
    path = path.partition("?")[0]
    path = path.partition("#")[0]

    if not path or path == ":memory:":
        return None

    return Path(path).expanduser()


def _integrity_check_pragma(mode: SqliteIntegrityCheckMode) -> str:
    if mode == SqliteIntegrityCheckMode.QUICK:
        return "PRAGMA quick_check;"
    return "PRAGMA integrity_check;"


def check_sqlite_integrity(
    path: Path,
    *,
    mode: SqliteIntegrityCheckMode = SqliteIntegrityCheckMode.FULL,
) -> IntegrityCheck:
    if not path.exists():
        return IntegrityCheck(ok=True, details=None)

    try:
        with sqlite3.connect(str(path)) as conn:
            cursor = conn.execute(_integrity_check_pragma(mode))
            rows = [row[0] for row in cursor.fetchall()]
    except sqlite3.DatabaseError as exc:
        return IntegrityCheck(ok=False, details=str(exc))

    if len(rows) == 1 and rows[0] == "ok":
        return IntegrityCheck(ok=True, details=None)

    if not rows:
        return IntegrityCheck(ok=False, details=f"{mode.value}_check returned no rows")

    details = "; ".join(str(row) for row in rows)
    return IntegrityCheck(ok=False, details=details)
